/***********************************************************************
 * This file is part of module_imagepreview.
 *
 * module_imagepreview is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_imagepreview is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_imagepreview.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include "imageviewmodel.h"

#include <QDebug>
#include <QModelIndex>
#include <QIcon>
#include <QPixmap>
#include <QBrush>
#include <QApplication>
#include <QPalette>
#include <QByteArray>
#include <QDataStream>

// we cannot load the file directly, the resources need to be loaded first
MiscTools::pQImage ImageViewModel::noThumb = _initNoThumb();
MiscTools::pQImage ImageViewModel::errorThumb = _initErrThumb();

MiscTools::pQImage ImageViewModel::_initNoThumb()
{
   Q_INIT_RESOURCE(infoimages);
   return MiscTools::pQImage(new QImage(tr(":/imagepreview/nothumb.png")));
}

MiscTools::pQImage ImageViewModel::_initErrThumb()
{
   Q_INIT_RESOURCE(infoimages);
   return MiscTools::pQImage(new QImage(tr(":/imagepreview/errorthumb.png")));
}

ImageViewModel::ImageViewModel(QObject *parent)
    : QAbstractTableModel(parent)
    , _currentEditing(-1)
{

}

int ImageViewModel::rowCount(const QModelIndex&) const
{
    return _images.count();
}

int ImageViewModel::columnCount(const QModelIndex& ) const
{
    return IMAGE_VIEW_COLUMNS;
}


QVariant ImageViewModel::data(const int row, const int col, int role) const
{

    const ImageData& i = * dynamic_cast<ImageData*>(_images[row].get());

    if (col > 1) {
        return QVariant();
    }

    if (Qt::CheckStateRole == role && col == IMAGE_VIEW_FILENAME_IDX)
    {
        return i.getUsed() ? Qt::Checked : Qt::Unchecked;

    } else if (role == Qt::TextAlignmentRole) {
        if (col == IMAGE_VIEW_FILENAME_IDX) {
            return int(Qt::AlignLeft | Qt::AlignVCenter);
        } else {
            return int(Qt::AlignCenter | Qt::AlignVCenter);
        }

    } else if (role == Qt::EditRole) {
        // not implemented as not needed

    } else if (role == Qt::BackgroundRole) {
        QBrush colorAlternate = QApplication::palette("tableView_Images")
                .brush(QPalette::Active,
                       row %2 == 0 ?
                       QPalette::Base : QPalette::AlternateBase);

        return colorAlternate;
    } else if (role == Qt::DecorationRole) {
        if (col == IMAGE_VIEW_PREVIEW_IDX) {
            if (i.hasThumbnail() &&
                i.getThumbnail()->size().width() > 0
                && i.getThumbnail()->size().height() > 0)
            {
                return QIcon(QPixmap::fromImage(*i.getThumbnail()));
            } else if (i.hasThumbnail() == true) {
                return QIcon(QPixmap::fromImage(*errorThumb));
            } else {
                return QIcon(QPixmap::fromImage(*noThumb));
            }
        }
    } else if (role == Qt::DisplayRole) {
        if (col == IMAGE_VIEW_FILENAME_IDX) {
            return i.getFileInfo().fileName();
        }
    }

    return QVariant();
}

QVariant ImageViewModel::data(const QModelIndex& index, int role) const
{
    if (!index.isValid()) {
        return QVariant();
    }

    return data(index.row(), index.column(), role);


}

QVariant ImageViewModel::headerData(int section,
                                    Qt::Orientation orientation,
                                    int role) const
{
    if (role == Qt::DisplayRole && orientation == Qt::Horizontal) {
        if (section == IMAGE_VIEW_PREVIEW_IDX)
            return tr("Thumbnail");
        else if (section == IMAGE_VIEW_FILENAME_IDX)
            return tr("Filename");
        else
            return QVariant();
    }
    return QVariant();
}

bool ImageViewModel::setData ( const QModelIndex & index,
                               const QVariant & value,
                               int role)
{
    if (role == Qt::CheckStateRole
        && index.column() == IMAGE_VIEW_FILENAME_IDX)
    {
        dynamic_cast<ImageData*>(_images[index.row()].get())->setUsed(value.toBool());
        _currentEditing = -1;
        return true;
    }


    _currentEditing = -1;
    return QAbstractTableModel::setData(index, value, role);
}

Qt::ItemFlags ImageViewModel::flags ( const QModelIndex & index ) const
{
    if ( index.column() == IMAGE_VIEW_FILENAME_IDX) {
        return Qt::ItemIsEnabled | Qt::ItemIsUserCheckable | Qt::ItemIsSelectable;
    } else {
        return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
    }
}

const QList<ImageDataPtr>& ImageViewModel::getList() const
{
    return _images;
}

void ImageViewModel::setList(const QList<ImageDataPtr>& newList)
{
    emit layoutAboutToBeChanged();
    _images.clear();
    for (int i=0; i<newList.size(); i++) {
        _images.push_back( ImageDataPtr(new ImageData(*newList[i])) );
    }
    emit layoutChanged();

    _updateViews();
}


void ImageViewModel::updateList(const int pos, const ImageData& im)
{
    if (pos >= _images.size() || pos < 0) {
        throw "wrong image index";
    }

    *(_images[pos]) = im;

    _updateViews();
}

void ImageViewModel::_updateViews(const int row)
{
    if (row < 0) {
        QModelIndex start = this->index(0,0);
        QModelIndex end = this->index(_images.size(), this->columnCount(start)-1);
        emit dataChanged(start, end);
    } else if (row != _currentEditing){
        // if pos was _currentEditing, current editing would be killed
        QModelIndex start = this->index(row,0);
        QModelIndex end = this->index(row, this->columnCount(start)-1);
        emit dataChanged(start, end);
    }
}

void ImageViewModel::selectAll()
{
    for (int i=0; i<_images.size(); i++) {
        _images[i]->setUsed(true);
    }
    _updateViews();
}

void ImageViewModel::deselectAll()
{
    for (int i=0; i<_images.size(); i++) {
        _images[i]->setUsed(false);
    }
    _updateViews();
}

void ImageViewModel::invertSelection()
{
    for (int i=0; i<_images.size(); i++) {
        _images[i]->setUsed(!_images[i]->getUsed()) ;
    }
    _updateViews();
}
