/***********************************************************************
 * This file is part of module_imageviewer.
 *
 * module_imageviewer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_imageviewer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_imageviewer.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QThread>
#include <QTimer>
#include <QDebug>
#include <QMediaPlayer>
#include <QApplication>

#ifdef _WIN32
#include <QVideoWidget>
#endif

#include "imagepreloader.h"

QMutex* ImagePreloader::changeLoaderConditionsMutex = NULL;

ImagePreloader::ImagePreloader()
    : _currentPosition(0)
    , _rawDisplayed(true)
    , _numThreads(QThread::idealThreadCount() == -1 ? 2 : QThread::idealThreadCount())
    , _player(new QMediaPlayer())
    , _videoOut(0)
{
    _loaderThreads.resize(_numThreads);
    for (int i = 0; i < _numThreads; i++) {
        _loaderThreads[i].thread = NULL;
    }

#ifdef _WIN32
        // windows needs something to render to
        // our "test" player will render on a video widget
        _videoOut = new QVideoWidget;
        _videoOut->setVisible(false);
        _player->setVideoOutput(_videoOut);
#endif

    // set number of resources to number of threads
    // this allows the main view window to wait
    // for urgently to be loaded images, as then there
    // is only one thread working
    //ImageViewerLoader::loadAll = new QSemaphore(_numThreads);
    changeLoaderConditionsMutex = new QMutex();

    connect(&_mediaTimer, SIGNAL(timeout()), SLOT(checkForVideo()));
}

ImagePreloader::~ImagePreloader()
{    
    // make sure threads are shut down
    // and deleted
    _deleteThreads(-1);
    delete changeLoaderConditionsMutex;
    changeLoaderConditionsMutex = 0;
    //delete ImageViewerLoader::loadAll;
}

void ImagePreloader::setRawFormats(QStringList rawFormats)
{
    _myRawFormats = rawFormats;
    for (int i=0; i<_rawFormats.size(); i++) {
        _myRawFormats[i] = _myRawFormats[i].toLower();
    }
}

ImageDataPtr ImagePreloader::getFileInfoPtr(const int position)
{
    if (position < 0 || position > _mapToOrig.size() || _mapToOrig[position] > _imagesOrig.size()) {
        return 0;
    }
    return _imagesOrig[_mapToOrig[position]].file;
}

ImageDataPtr ImagePreloader::getFileInfoPtrOverallList(const int position)
{
    if (position < 0 || position > _imagesOrig.size()) {
        return 0;
    }
    return _imagesOrig[position].file;
}

void ImagePreloader::_deleteThreads(const int num,
                                    const bool resetLoadingStat)
{    
    // make sure threads are shut down and deleted
    if (num >=0) {
        // delete only one thread
        if (num >= _numThreads) {
            return;
        }
        // first make sure there is no string left
        // nor any signals connected
        // particular image was loading so set back to not loaded
        if (resetLoadingStat
                && _loaderThreads[num].mapToPosition >= 0
                && _loaderThreads[num].mapToPosition < _imagesOrig.size()) {
            _imagesOrig[_loaderThreads[num].mapToPosition].status = LoadingStatusNotLoaded;
        }

        _loaderThreads[num].canonicalFilepath = "";
        _loaderThreads[num].mapToPosition = -1;

        if (_loaderThreads[num].thread != NULL) {
            ImageViewerLoader * thread = _loaderThreads[num].thread;
            //        disconnect(_loaderThreads[num].thread,
            //                   SIGNAL(doneLoading(const pQImage&, const QString&)),
            //                   this, SLOT(onImageLoaded(const pQImage&,const QString&)));

            // termination seems not to work properly ?!
            // anyways, decrease priority, such that they don't block up
            // too much of the required resources
            if (thread->isRunning()) {
                //thread->setPriority(QThread::LowestPriority);
                try {
                    thread->quit(); // may not be useful
                    thread->deleteLater();
                    //thread->terminate();
                    //thread->wait();

                    // in contrast to terminate, this will delete
                    // the thread when done
                    //thread->deleteInstance();
                } catch (...) {
                    qCritical() << "caught loader thread termination error";
                }
            } else {
                thread->deleteLater();
            }
            _loaderThreads[num].thread = NULL;
        }
    } else {
        // delete all threads recursively
        for (int i = 0; i < (int)_loaderThreads.size(); i++) {
            _deleteThreads(i, resetLoadingStat);
        }
    }
}

int ImagePreloader::setImages(const QList<ImageDataPtr>& newImages,
                              const int positionOrig,
                              int& positionNew)
{
    QMutexLocker l(this->changeLoaderConditionsMutex);

    // terminate running threads, without waiting
    _deleteThreads(-1, false);

    // copy new given list to this instance while checking if some
    // of these images have already been loaded for the currently
    // given list
    QList<LoadImageInfo> newListTmp;
    for (int i=0; i<(int) newImages.size(); i++) {
        LoadImageInfo tmp = LoadImageInfo(newImages[i]);

        // check if it has been loaded already, if yes copy pixmap
        // otherwise just add
        /*if (_imagesOrig.contains(tmp)) {
            newListTmp.push_back( _imagesOrig[_imagesOrig.indexOf(tmp)]);
            if (newListTmp.last().status != LoadingStatusIsLoaded) {
                newListTmp.last().status = LoadingStatusNotLoaded;
            }
        } else {*/
        newListTmp.push_back( tmp );
        //}
    }

    _imagesOrig.clear();
    _imagesOrig = newListTmp;

    positionNew = _createImageMap(positionOrig >= 0 && positionOrig < newImages.size() ? positionOrig : 0);

    return _mapToOrig.size();
}

int ImagePreloader::setRawDisplayed(const bool rawDisp,
                                    int& newNumImages)
{
    if (_rawDisplayed == rawDisp) {
        // nothing has changed
        return -1;
    }

    {
        QMutexLocker l(this->changeLoaderConditionsMutex);
        _rawDisplayed = rawDisp;

        // empty anyways ?
        if (_currentPosition < 0
                || _currentPosition >= _imagesOrig.size()
                || _mapToOrig.size() == 0) {
            return -1;
        }
        // find position in raw image
        const int positionRaw = _mapToOrig[_currentPosition];
        // create new map with that position
        _createImageMap(positionRaw);
    }
    _preLoadList.clear();

    // force new loading steps for threads
    onImageLoaded(MiscTools::pQImage(), "");

    newNumImages = _mapToOrig.size();
    return _currentPosition;
}


int ImagePreloader::_createImageMap(const int desiredPositionOrig)
{
    int desiredPosition = desiredPositionOrig;
    _mapToOrig.clear();
    _currentPosition = 0;


    if (_rawDisplayed) {
        // all raw images are displayed, so
        // we create an identity map
        _currentPosition = desiredPosition;
        _mapToOrig.resize(_imagesOrig.size());
        for (int i=0; i<_mapToOrig.size(); i++) {
            _mapToOrig[i] = i;
        }
    } else {
        // do not show raw images, if there's a corresponding
        // (same basename) file which is not of raw format
        const int s = (int)_imagesOrig.size();
        for (int i=0; i<s; i++) {
            const QString basename = _imagesOrig[i].file->getFileInfo().completeBaseName();
            const QString ext = _imagesOrig[i].file->getFileInfo().suffix().toLower();
            const QString path = _imagesOrig[i].file->getFileInfo().canonicalPath();

            if (_myRawFormats.contains(ext)) {
                // if it is indeed a raw file, then try to find if there exists a
                // corresponding jpg file. if not DO add raw file
                bool foundOther = false;
                for (int j=0; j<s; j++) {
                    if (j == i) {
                        continue;
                    }
                    if (_imagesOrig[j].file->getFileInfo().completeBaseName() == basename
                            && _imagesOrig[j].file->getFileInfo().canonicalPath() == path) {
                        const QString ext2 = _imagesOrig[j].file->getFileInfo().suffix().toLower();

                        if (!_myRawFormats.contains(ext2)) {
                            // found another file
                            foundOther = true;

                            // if the current raw file happened to be wished to be displayed
                            // change desired position to the other, not-raw file
                            if (i == desiredPosition) {
                                desiredPosition = j;
                            }
                            break;
                        }
                    }
                }
                if (!foundOther) {
                    _mapToOrig.append(i);
                }
            } else {
                // is no raw file, so simply append
                _mapToOrig.append(i);
            }
        }

        // find corrected current position,
        // this does apply for the position shift
        // and the change of file if desiredposition was
        // a raw file
        for (int i=0; i<_mapToOrig.size(); i++) {
            if (_mapToOrig[i] == desiredPosition) {
                _currentPosition = i;
            }
        }
    }

    return _currentPosition;

}

void ImagePreloader::_newThread(const QString& canonicalFilepath,
                                const int num,
                                const int position)
{
    if (_loaderThreads[num].thread != NULL) {
        _deleteThreads(num, false);
    }
    // start new thread
    _loaderThreads[num].canonicalFilepath = canonicalFilepath;
    _loaderThreads[num].thread = new ImageViewerLoader(_loaderThreads[num].canonicalFilepath);
    if (_loaderThreads[num].thread) {
        connect (_loaderThreads[num].thread,
                 SIGNAL(doneLoading(const MiscTools::pQImage&, const QString&)),
                 this,
                 SLOT(onImageLoaded(const MiscTools::pQImage&, const QString&)),
                 Qt::QueuedConnection);
        _imagesOrig[_mapToOrig[position]].status = LoadingStatusIsLoading;
        _loaderThreads[num].thread->start();

        _loaderThreads[num].mapToPosition = _mapToOrig[position];
    } else {
        qCritical() << "could not create thread";
    }
}

void ImagePreloader::checkForVideo()
{
    bool newData = false;
    if (_player->mediaStatus() != QMediaPlayer::NoMedia) {
        if (_player->mediaStatus() == QMediaPlayer::LoadingMedia) {
            _mediaTimer.stop();
            _mediaTimer.setSingleShot(true);
            _mediaTimer.setInterval(100);
            _mediaTimer.start();
        } else {
            // find video again
            const QFileInfo test(_player->media().canonicalUrl().toLocalFile());
            for (int j=0; j<(int) _mapToOrig.size(); j++) {
                const int posOrig = _mapToOrig[j];
                if (_imagesOrig[posOrig].file->getFileInfo() ==  test) {
                    if (_player->mediaStatus() == QMediaPlayer::InvalidMedia) {
                        _imagesOrig[posOrig].status = LoadingStatusFailedLoading;
                        emit errorLoading(test.canonicalFilePath());
                    } else {
                        _imagesOrig[posOrig].status = LoadingStatusIsVideo;
                        emit doneLoading(test.canonicalFilePath());
                    }
                    break;
                }
            }
            _player->stop();
            newData = true;
        }
    }

    if (!_playlist.isEmpty() &&
            (_player->mediaStatus() == QMediaPlayer::NoMedia || newData) ) {
        _player->setMedia(QUrl::fromLocalFile(_playlist.first()));
        _playlist.pop_front();
        _mediaTimer.stop();
        _mediaTimer.setSingleShot(true);
        _mediaTimer.setInterval(100);
        _mediaTimer.start();
    }
}

void ImagePreloader::onImageLoaded(const MiscTools::pQImage image,
                                   const QString& canonicalFilePath)
{
    MiscTools::pQPixmap pixmap (QPixmap::fromImage(*image));
    QMutexLocker l(this->changeLoaderConditionsMutex);
    // update respective file
    for (int j=0; j<(int) _mapToOrig.size(); j++) {
        const int posOrig = _mapToOrig[j];
        if (posOrig < 0 || posOrig >= _imagesOrig.size()) {
            continue;
        }
        if (_imagesOrig[posOrig].file->getFileInfo().canonicalFilePath() ==  canonicalFilePath) {
            if (pixmap->width() > 0 && pixmap->height() > 0) {
                _imagesOrig[posOrig].status = LoadingStatusIsLoaded;
            } else if (_imagesOrig[posOrig].status != LoadingStatusIsTestedForVideo) {
                _imagesOrig[posOrig].status = LoadingStatusIsTestedForVideo;
                // test if it is a video
                _playlist.append(canonicalFilePath);
                _mediaTimer.stop();
                _mediaTimer.setSingleShot(true);
                _mediaTimer.setInterval(100);
                _mediaTimer.start();
                break;
            }


            int threadNo = -1;
            for (int i=0; i<_numThreads; i++) {
                if (_loaderThreads[i].canonicalFilepath == canonicalFilePath) {
                    threadNo = i;
                    break;
                }
            }

            // adapt to possible inherent orientation of the image
            const QString suff = _imagesOrig[posOrig].file->getFileInfo().suffix().toLower();
            const int diffRot = _imagesOrig[posOrig].addDesiredRotation;
            // make sure the original transform already cares
            // for user desired rotation (i.e. if still loading and the
            // user has already pressed "rotate"):
            QTransform transform = QTransform().rotate(diffRot);
            if (suff == "jpg" || suff == "jpeg" ) {
                switch (_imagesOrig[posOrig].file->getOrientation()) {
                case MiscTools::OrientationJPG_Rot180:
                    *_imagesOrig[posOrig].pixmap = pixmap->transformed(transform.rotate(180));
                    break;
                case MiscTools::OrientationJPG_rotLeft:
                    *_imagesOrig[posOrig].pixmap = pixmap->transformed(transform.rotate(90));
                    break;
                case MiscTools::OrientationJPG_rotRight:
                    *_imagesOrig[posOrig].pixmap = pixmap->transformed(transform.rotate(-90));
                    break;
                case MiscTools::OrientationJPG_flippedH:
                case MiscTools::OrientationJPG_flippedV:
                case MiscTools::OrientationJPG_flippedHRotLeft:
                case MiscTools::OrientationJPG_flippedHRotRight:
                    // flipping of qpixmap seems not as trivial
                    // to perform efficiently
                    // (e.g. conversion to QImage etc)
                    qWarning() << "orientation tags with flipped images"
                             << " not supported yet";
                case MiscTools::OrientationJPG_normal:
                default:
                    *_imagesOrig[posOrig].pixmap = pixmap->transformed(transform);
                    break;
                }
            } else {
                *_imagesOrig[posOrig].pixmap = pixmap->transformed(transform);;
            }

            _deleteThreads(threadNo, false);

            emit doneLoading(canonicalFilePath);
            break;
        }
    }

    QTimer::singleShot(0, this, SLOT(_setupNewThreads()));
}

void ImagePreloader::_setupNewThreads()
{

    // cleanup (remove already loaded elements from loading queue)
    for (int i=_preLoadList.size()-1; i>=0; i--) {
        if (_imagesOrig[_mapToOrig[_preLoadList[i].first]].status == LoadingStatusIsLoaded
                || _imagesOrig[_mapToOrig[_preLoadList[i].first]].status == LoadingStatusIsVideo
                || _imagesOrig[_mapToOrig[_preLoadList[i].first]].status == LoadingStatusIsTestedForVideo
                || _imagesOrig[_mapToOrig[_preLoadList[i].first]].status == LoadingStatusFailedLoading) {
            _preLoadList.remove(i);
        }
    }

    // check for high priority
    bool foundPriorityThread = false;
    for (int i=0; i < _preLoadList.size(); i++) {
        if (_preLoadList[i].second == true) {
            const int pos = _preLoadList[i].first;
            const QString filepath = _imagesOrig[_mapToOrig[pos]].file->getFileInfo().canonicalFilePath();

            _checkIfLoadingOrNewHighPriority(filepath, pos);
            foundPriorityThread = true;
            _preLoadList.remove(i);
            break;
        }
    }

    // otherwise check for next/previous/first/last elements to be loaded
    for (int i=0; i<NUM_IMAGES_TO_EACH_SIDE; i++) {
        //end
        int p = i + (_mapToOrig.size() - NUM_IMAGES_TO_EACH_SIDE) ;
        if ( p >=0 ) {
            if (_imagesOrig[_mapToOrig[p]].status == LoadingStatusNotLoaded) {
                _preLoadList.push_front(QPair<int, bool>(p, false));
            }
        }
        // start
        p = (NUM_IMAGES_TO_EACH_SIDE-1) - i;
        if (p < _mapToOrig.size()) {
            if (_imagesOrig[_mapToOrig[p]].status == LoadingStatusNotLoaded) {
                _preLoadList.push_front(QPair<int, bool>(p, false));
            }
        }
        // previous
        p = _currentPosition-(NUM_IMAGES_TO_EACH_SIDE-i);
        if ( p >=0 && p <_mapToOrig.size()) {
            if (_imagesOrig[_mapToOrig[p]].status == LoadingStatusNotLoaded) {
                _preLoadList.push_front(QPair<int, bool>(p, false));
            }
        }
        // following
        p = _currentPosition+(NUM_IMAGES_TO_EACH_SIDE-i);
        if ( p >=0 && p <_mapToOrig.size()) {
            if (_imagesOrig[_mapToOrig[p]].status == LoadingStatusNotLoaded) {
                _preLoadList.push_front(QPair<int, bool>(p, false));
            }
        }
    }

    // check for out of range, to delete pixmaps
    const int s = _mapToOrig.size();
    const int buff = 5;
    for (int i=0; i<s; i++) {
        if ( abs(i -s) > buff * NUM_IMAGES_TO_EACH_SIDE
             && i > buff * NUM_IMAGES_TO_EACH_SIDE
             && abs(i-_currentPosition) > buff * NUM_IMAGES_TO_EACH_SIDE) {
            if (_imagesOrig[_mapToOrig[i]].status == LoadingStatusIsLoaded) {
                _imagesOrig[_mapToOrig[i]].status = LoadingStatusNotLoaded;
                *_imagesOrig[_mapToOrig[i]].pixmap = QPixmap();
            }
        }
    }

    // setup new threads
    if (_preLoadList.size() > 0 && !foundPriorityThread ) {
        for (int i=0; i < _numThreads; i++) {
            if (_loaderThreads[i].thread == NULL) {
                // make sure, it's not already being loaded
                int pos = -1;
                bool isComputing = true;
                QString newThreadFilePath = "";
                while (isComputing && _preLoadList.size() > 0) {
                    pos = _preLoadList[0].first;
                    newThreadFilePath =
                            _imagesOrig[_mapToOrig[pos]].file->getFileInfo().canonicalFilePath();
                    isComputing = false;
                    for (int j=0; j < _numThreads; j++) {
                        if (_loaderThreads[j].canonicalFilepath == newThreadFilePath) {
                            isComputing = true;
                            break;
                        }
                    }
                    if (isComputing) {
                        _preLoadList.pop_front();
                    }
                }
                if (_preLoadList.size() == 0) {
                    break;
                }
                _newThread(newThreadFilePath, i, pos);
            }
        }
    }    

}

void ImagePreloader::_checkIfLoadingOrNewHighPriority(const QString& filepath,
                                                      const int position)
{
    // the desired pixmap has not been loaded yet
    // first of all check, if it is already being loaded by a thread
    bool foundIsAlreadyLoading = false;
    for (int i=0; i < _numThreads; i++) {
        if (_loaderThreads[i].thread == NULL) {
            continue;
        }
        if (_loaderThreads[i].canonicalFilepath == filepath) {
            // image is already being loaded by one of the threads
            foundIsAlreadyLoading = true;
            // QThread::TimeCriticalPriority is max, but probably not worth it
            if (_loaderThreads[i].thread->isRunning()) {
                _loaderThreads[i].thread->setPriority(QThread::HighestPriority);
            }

            // no pause other threads, we most urgently need this one image now.
            // however, pausing not quite supported ?
            // so kill other threads instead
            for (int j=0; j< _numThreads; j++) {
                if (j != i && _loaderThreads[j].thread != NULL) {
                    _deleteThreads(j, true);
                }
            }
            break;
        }
    }
    if (!foundIsAlreadyLoading) {
        // none of the threads is already computing this particular image
        // so start a thread and pause all other threads
        // however, pausing not quite supported ?
        // so kill other threads instead
        _deleteThreads(-1, true);

        // start new thread
        _newThread(filepath, 0, position);
    }
}

bool ImagePreloader::isVideoAtPosition(const int position) const
{
    return (!(position < 0 || position >= _mapToOrig.size()
               || _mapToOrig[position] < 0 || _mapToOrig[position]  >= _imagesOrig.size())
            && (_imagesOrig[_mapToOrig[position]].status == LoadingStatusIsVideo) );
}

MiscTools::pQPixmap ImagePreloader::getImageAtPosition(const int position,
                                           bool& found,
                                           QString& filepath,
                                           bool& isloading,
                                           bool& isVideo)
{
    QMutexLocker l(this->changeLoaderConditionsMutex);
    found = true;

    if (position < 0 || position >= _mapToOrig.size()
            || _mapToOrig[position] < 0 || _mapToOrig[position]  >= _imagesOrig.size()) {
        found = false;
        filepath = "";
        return MiscTools::pQPixmap();
    }

    isloading = _imagesOrig[_mapToOrig[position]].status == LoadingStatusIsLoading || _imagesOrig[_mapToOrig[position]].status == LoadingStatusNotLoaded;

    filepath = _imagesOrig[_mapToOrig[position]].file->getFileInfo().canonicalFilePath();
    if (_imagesOrig[_mapToOrig[position]].status == LoadingStatusIsLoaded) {
        found = true;
        _currentPosition = position;
        QTimer::singleShot(0, this, SLOT(_setupNewThreads()));
        if (_imagesOrig[_mapToOrig[position]].pixmap != 0) {
            return *_imagesOrig[_mapToOrig[position]].pixmap;
        } else {
            return MiscTools::pQPixmap();
        }
    } else if (_imagesOrig[_mapToOrig[position]].status == LoadingStatusIsVideo) {
        found = false;
        _currentPosition = position;
        return MiscTools::pQPixmap();
    } else if (_imagesOrig[_mapToOrig[position]].status == LoadingStatusFailedLoading) {
        found = false;
        // try to again load the file, but only if it hasn't been tried just ago
        if (_currentPosition != position) {
            _currentPosition = position;
            QTimer::singleShot(0, this, SLOT(_setupNewThreads()));
        }
        _currentPosition = position;
        return MiscTools::pQPixmap();
    } else if (_imagesOrig[_mapToOrig[position]].status == LoadingStatusIsTestedForVideo) {
        isVideo = true;
        found = false;
        return MiscTools::pQPixmap();
    } else {
        // the desired pixmap has not been loaded yet
        // first of all check, if it is already being loaded by a thread
        const QString filepath = _imagesOrig[_mapToOrig[position]].file->getFileInfo().canonicalFilePath();
        _checkIfLoadingOrNewHighPriority(filepath, position);

        // next put up high priority loading of next image (in direction)
        if (_currentPosition <= position) {
            if (position == _mapToOrig.size() -1) {
                if (position > 0) {
                    _preLoadList.push_front(QPair<int, bool>(position-1, true));
                }
            } else {
                _preLoadList.push_front(QPair<int, bool>(position+1, true));
            }
        } else {
            if (position == 0) {
                if (position < _mapToOrig.size() -1) {
                    _preLoadList.push_front(QPair<int, bool>(1, true));
                }
            } else {
                _preLoadList.push_front(QPair<int, bool>(position-1, true));
            }
        }

        // adapt position
        _currentPosition = position;
        found = false;
    }    

    return MiscTools::pQPixmap();
}


void ImagePreloader::clearImages()
{
    _deleteThreads(-1, false);
    _imagesOrig.clear();
    _currentPosition=0;
    _preLoadList.clear();
    _mapToOrig.clear();
}

void ImagePreloader::setDesiredOrientation(const int position,
                            const int addRotation)
{
    if (position < 0 || position >= _mapToOrig.size()) {
        qCritical() << "invalid index for rotation";
        return;
    }
    const int posOrig = _mapToOrig[position];
    _imagesOrig[posOrig].addDesiredRotation = (_imagesOrig[posOrig].addDesiredRotation + addRotation) % 360;
    if (_imagesOrig[posOrig].status == LoadingStatusIsLoaded) {
        QTransform transform;
        *_imagesOrig[posOrig].pixmap = _imagesOrig[posOrig].pixmap->transformed(transform.rotate(addRotation));
    }
}
