/***********************************************************************
 * This file is part of module_imageviewer.
 *
 * module_imageviewer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_imageviewer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_imageviewer.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#ifndef IMAGEPRELOADER_H
#define IMAGEPRELOADER_H

#include <QMutex>
#include <QStringList>
#include <QPair>
#include <QTimer>
#include <QtMultimedia/QMediaPlayer>
#include <QtMultimediaWidgets/QVideoWidget>

#include "imageviewerloader.h"
#include "imagedata.h"
#include "threadinfo.h"
#include "loadingstatus.h"
#include "loadimageinfo.h"
#include "smartpointers.h"

#define NUM_IMAGES_TO_EACH_SIDE 3

class QMediaPlayer;

class ImagePreloader : public QObject
{
    Q_OBJECT
public:
    ImagePreloader();
    ~ImagePreloader();

    int setImages(const QList<ImageDataPtr>& newImages,
                  const int position,
                  int &positionNew);

    /**
     * return pixmap for image at position
     * @param position: current desired position
     * @return bool found if false, it is still loading
     * @return QString filepath the filepath of the desired image
     * @return QPixmap image pixmap or empty pixmap if
     * loading failed
     */
    MiscTools::pQPixmap getImageAtPosition(const int position, bool &found, QString &filepath, bool &isloading, bool &isVideo);
    bool isVideoAtPosition(const int position) const;

    int setRawDisplayed(const bool rawDisp, int &newNumImages);

    void setRawFormats(QStringList rawFormats);

    static QMutex* changeLoaderConditionsMutex;

    int getNumThreads() const { return _numThreads; }

    void clearImages();

    ImageDataPtr getFileInfoPtr(const int position);
    int getNumFilesOverall() const { return _imagesOrig.size(); }
    ImageDataPtr getFileInfoPtrOverallList(const int position);

    void setDesiredOrientation(const int position,
                               const int addRotation);

signals:
    void doneLoading(const QString&);
    void errorLoading(const QString&);

private slots:
    void onImageLoaded(const MiscTools::pQImage, const QString& );

    void _setupNewThreads();

    void checkForVideo();

private:
    /**
     * Create a map of indexes to the original images
     * The map respects the mode of display of raw images
     *
     * @param int desiredPosition the desired position in the original image
     */
    int _createImageMap(const int desiredPositionOrig);

    void _deleteThreads(const int num, const bool resetLoadingStat = false);

    void _newThread(const QString& canonicalFilepath,
                    const int num,
                    const int position);

    void _checkIfLoadingOrNewHighPriority(const QString& filepath,
                                          const int position);


    int _currentPosition;
    QList<LoadImageInfo> _imagesOrig;

    QVector<int> _mapToOrig;
    QVector<int> _mapToOrigByFilename;

    bool _rawDisplayed;

    QVector<ThreadInfo> _loaderThreads;

    const int _numThreads;

    QStringList _myRawFormats;

    /**
     * first: index
     * second: priority, false normal, true only load this
     */
    QVector<QPair<int, bool> > _preLoadList;

    QMediaPlayer * _player;
    QVideoWidget * _videoOut;
    QStringList _playlist;
    QTimer _mediaTimer;
};

#endif // IMAGEPRELOADER_H
