/***********************************************************************
 * This file is part of module_misc.
 *
 * module_misc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_misc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_misc.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QStringList>
#include <QFileInfo>
#include <QDebug>
#include <QMultiMap>
#include <QVector>
#include <QPair>

#include "filenamenumbering.h"
#include "misctools.h"

namespace MiscTools
{

/**
  * a separator to build up a filenamestring
  */
#ifdef _WIN32
    #define SEPARATOR QString("<")
#else
    #define SEPARATOR QString('/')
#endif



QVector<FilenameNumbering::DT> FilenameNumbering::getExistingFileTimeForCopyMerge(
        const QStringList& filenames,
        const QString& taggedFileName,
        const QVector<QPair<QString, QString> >& tagsAndDateMap,
        QString& datePattern)
{
    QMultiMap<int, int> tagPositions;
    QVector<QVector<QVariant> > filenameNumericValues;
    QVector<int> datePatternsUsed;
    QVector<QVector< QPair<int, int> > > tagSourceFNPositionsAndLengths;
    _getTagMap(filenames,
               taggedFileName,
               tagsAndDateMap,
               tagPositions,
               datePatternsUsed,
               datePattern,
               filenameNumericValues,
               tagSourceFNPositionsAndLengths);



    // now, check all filenames for given dates etc
    const int n = filenames.size();
    QVector<DT> fileTimesFromFilename;
    if (n !=filenameNumericValues.size()) {
        qWarning() << "number of filenames mismatch determined number of numeric values";
        return QVector<FilenameNumbering::DT> ();
    }
    for (int i=0; i<n; i++) {
        const bool foundFilenameTime = filenameNumericValues[i].size() != 0;

        QString currDateString;
        QDateTime currDateTime;
        if (foundFilenameTime) {
            for (int j=0; j<datePatternsUsed.size(); j++) {
                if (filenameNumericValues[i][datePatternsUsed[j]].type() != QVariant::Int) {
                    qCritical() << "tried to use string as running number, aborting";
                    fileTimesFromFilename.clear();
                    return fileTimesFromFilename;
                }
                const int val = filenameNumericValues[i][datePatternsUsed[j]].toInt();
                currDateString += QString(" %1").arg(val);
            }

            currDateTime = QDateTime::fromString(currDateString, datePattern);
        } else {
            qWarning() << "could not obtain file time from filename, trying exif date now..";
        }

        // now also check, if it is possible to determine EXIF time
        QDateTime exifTime;
        OrientationJPG tmp;
        bool exact = false;
        const bool foundExif = MiscTools::getTimeStampAndOrientation(filenames[i],
                                                                     exifTime,
                                                                     tmp);
        if (foundExif) {
            // check if filename datetime hint is consistent with real
            // exif time
            if (!foundFilenameTime || exifTime.toString(datePattern) == currDateString) {
                // ok, so now choose exiftime as datetime
                // since it might be more accurate
                // (mostly filename hints for date/time consist of
                //  year, month, day, only)
                currDateTime = exifTime;
                exact = true;
            }
        }

        fileTimesFromFilename.push_back(DT(currDateTime, exact));
    }

    if (fileTimesFromFilename.size() == filenames.size()) {
        return fileTimesFromFilename;
    } else {
        return QVector<FilenameNumbering::DT>();
    }
}

bool FilenameNumbering::mergeAndReorderFiles(
        const QStringList& existingFiles,
        const QVector<QDateTime>& newFileTimes,
        const QString& taggedFileName,
        const QVector<QPair<QString, QString> >& tagsAndDateMap,
        QVector<QPair<int,int> >& newlySortedList,
        QVector<FilenameNumbering::DT>& fileDatesExisting)
{
    // try to determine
    QString datePattern;
    fileDatesExisting = getExistingFileTimeForCopyMerge(
                existingFiles,
                taggedFileName,
                tagsAndDateMap,
                datePattern);

    const int nE = existingFiles.size();
    if (nE != fileDatesExisting.size()) {
        qCritical() << "cannot merge new file times to existing files, error occured";
        return false;
    }
    const int nN = newFileTimes.size();

    // sort into new list, using mergesort
    // - if existing file has no accurate time
    //   put new file with same datestring behind
    // - else if existing file has accurate time
    //   put new file relative to existing file
    //
    // this allows for approximate but accurate
    // insertion of new files into existing files
    newlySortedList.clear();
    newlySortedList.reserve(nN + nE);
    int posE=0, posN=0;
    while (newlySortedList.size() < (nE + nN) ) {
        if (posE == nE) {
            // now copy all new files to new list
            for ( ; posN<nN; posN++) {
                newlySortedList.push_back(QPair<int,int>(-1, posN));
            }
        } else if (posN == nN) {
            // now copy all existing files to new list
            for ( ; posE<nE; posE++) {
                newlySortedList.push_back(QPair<int,int>(posE, -1));
            }

        } else if (fileDatesExisting[posE].second == false) {
            // comparison with reduced accurately new time
            const QDateTime newFileReducedTime =
                    QDateTime::fromString(newFileTimes[posN].toString(datePattern), datePattern);
            if (fileDatesExisting[posE].first <= newFileReducedTime) {
                newlySortedList.push_back(QPair<int,int>(posE, -1));
                posE++;
            } else {
                newlySortedList.push_back(QPair<int,int>(-1, posN));
                posN++;
            }
        } else {
            // proper comparison, as the existing file-time could be
            // determined exactly (i.e. EXIF)
            if (fileDatesExisting[posE].first <= newFileTimes[posN]) {
                newlySortedList.push_back(QPair<int,int>(posE, -1));
                posE++;
            } else {
                newlySortedList.push_back(QPair<int,int>(-1, posN));
                posN++;
            }
        }
    }

    return true;
}


} // end namespace MiscTools
