/***********************************************************************
 * This file is part of module_misc.
 *
 * module_misc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_misc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_misc.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/


#include <QMatrix>
#include <QPropertyAnimation>
#include <qmath.h>
#include <QTimer>
#include <QPainter>

#include "imageflip.h"
#include "smartpointers.h"

namespace MiscTools
{

ImageFlip::ImageFlip(QWidget *parent)
    : QLabel(parent)
    , _currentStatus(0)
    , _viewSize(128)
    , _animation(NULL)
    , _currEasingCurve(QEasingCurve::Linear)
    , _time(1500)
{
    setupAnimation();
}

ImageFlip::ImageFlip(const pQImage& im, QWidget *parent)
    : QLabel(parent)
    , _currentStatus(0)
    , _viewSize(128)
    , _animation(NULL)
    , _currEasingCurve(QEasingCurve::Linear)
    , _time(1500)
{

    setCurrentStatus(_currentStatus);

    setupAnimation();

    setNewImage(im);
}
ImageFlip::~ImageFlip()
{
    delete _animation;
    _animation = 0;
}

void ImageFlip::setupAnimation()
{
    _animation = new QPropertyAnimation(this, "currentStatus");
    _animation->setDuration(_time);
    _animation->setStartValue(0);
    _animation->setEndValue(1);
    _animation->setLoopCount(1);
}

void ImageFlip::startAnimation()
{
    if (_animation->state() == QAbstractAnimation::Running) {
        _animation->stop();
    }
    _animation->start(QAbstractAnimation::KeepWhenStopped);
}

void ImageFlip::setAnimationMode(const QEasingCurve curve, const qreal amplitude)
{
    _currEasingCurve = curve;
    _currEasingCurve.setAmplitude(amplitude);
    _animation->setEasingCurve(_currEasingCurve);
}

void ImageFlip::setTime(const qint32 time)
{
    _time = time;
    _animation->setDuration(_time);
}

qreal ImageFlip::currentStatus() const
{
    return _currentStatus;
}

void ImageFlip::setNewImage(const pQImage &im)
{
    _newOrig = im->scaled(QSize(_viewSize,_viewSize), Qt::KeepAspectRatio );
    //startAnimation();
    QTimer::singleShot(0, this, SLOT(startAnimation()));
}

void ImageFlip::setCurrentStatus(const qreal status)
{
    pQImage resultImage(new QImage(_viewSize, _viewSize, QImage::Format_ARGB32_Premultiplied));
    const QSize tS = QSize(_viewSize,_viewSize);
    const QSize dS = 0.5*(tS - _newOrig->size());
    const QSize sS = 0.5*(tS - _orig->size());

    QPainter painter(resultImage.get());
    painter.setCompositionMode(QPainter::CompositionMode_Source);
    painter.fillRect(resultImage->rect(), Qt::transparent);
    painter.drawImage(dS.width(), dS.height() - (1.0-status) * float(_viewSize), *_newOrig);
    painter.drawImage(sS.width(), sS.height() + status * float(_viewSize), *_orig);
    painter.end();

    this->setPixmap(QPixmap::fromImage(*resultImage));
    _currentStatus = status;
    if (status == 1.0) {
        _orig = _newOrig;
    }
}

void ImageFlip::setSize(const int viewSize)
{
    _viewSize = viewSize;
}

QSize ImageFlip::minimumSizeHint () const
{
    const int realSize = _viewSize + ADD_BORDER_TO_IMAGE;
    QSize out(realSize, realSize);
    return out;
}

QSize ImageFlip::sizeHint () const
{
    const int realSize = _viewSize + ADD_BORDER_TO_IMAGE;
    QSize out(realSize, realSize);
    return out;
}

} // end namespace MiscTools
