/***********************************************************************
 * This file is part of module_misc.
 *
 * module_misc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_misc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_misc.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QDir>
#include <QDebug>
#include <QCryptographicHash>
#include <QDateTime>
#include <QString>
#include <QVector>
#include <QLocale>
#include <QSettings>
#include <QDesktopWidget>

#include "misctools.h"
#include "FreeImage.h"
#include <cassert>
#include <exception>

#include <stdio.h>
#include <libexif/exif-loader.h>
#include <libexif/exif-data.h>
#include <FreeImage.h>

extern "C" {
#include "jpeg-data.h"
}

#ifdef _WIN32
#include <qstring.h>
#endif

namespace MiscTools
{

#ifdef _WIN32
QString MiscTools::getDevicePath(const QString& deviceID,
								 const QString& number)
{
    QSettings settings(QString("HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Control\\Class\\"
                               "%1\\%2\\DeviceData").arg(deviceID).arg(number),
						QSettings::NativeFormat);
    QString out = settings.value("WpdDevicePnPID", "").toString();
	return out;
}
#endif

bool MiscTools::_dialogLock = false;

bool MiscTools::lockDialog()
{
    if (_dialogLock) {
        return false;
    }
    _dialogLock = true;
    return true;
}

void MiscTools::unlockDialog()
{
    _dialogLock = false;
}

QRect MiscTools::retrieveDesktopResolution()
{
    QDesktopWidget d;

    // multi-screen support
    QRect s = d.screenGeometry(d.screenNumber());
    return s;
}

QString MiscTools::appDataPath()
{
#ifdef _WIN32
    QString out = QDir::homePath();
    QDir search(out);

    if (search.exists(QDir::toNativeSeparators("AppData/Roaming"))) {
        out += "/Appdata/Roaming";
    } else if (search.exists(QDir::toNativeSeparators("AppData/Local"))) {
        out += "/Appdata/Local";
    } else if (search.exists(QDir::toNativeSeparators("Application Data"))) {
        out += "/Application Data";
    }
    //return QDir::toNativeSeparators(out + QDir::toNativeSeparators("/Application Data/"));
    return out;
#else
    return QDir::toNativeSeparators(QDir::homePath());
#endif
}

pQImage MiscTools::loadImageFreeImage(const QString& filename)
{
    pQImage out;

    FIBITMAP *bitmap = NULL;
    try {
        // load a file using the FIF_LOAD_NOPIXELS flag
        FREE_IMAGE_FORMAT fif = FreeImage_GetFIFFromFilename(filename.toLocal8Bit().data());
        bitmap = FreeImage_Load(fif, filename.toLocal8Bit().data());

        if (bitmap) {
            FIBITMAP *src = FreeImage_ConvertTo32Bits(bitmap);
            FreeImage_Unload(bitmap);
            if (src) {

                //When the source bitmap uses a 32-bit padding, you can calculate the pitch using the
                //following formula:
                //int pitch = ((((bpp * width) + 31) / 32) * 4);

                const int scan_width = FreeImage_GetPitch(src);
                const int width = FreeImage_GetWidth(src);
                const int height = FreeImage_GetHeight(src);
                BYTE *bits = (BYTE*)malloc(sizeof(BYTE)*height * scan_width);

                if ( height == 0) {
                    return QImage();
                }

                FreeImage_ConvertToRawBits(bits,
                                           src,
                                           scan_width,
                                           32,
                                           FI_RGBA_RED_MASK,
                                           FI_RGBA_GREEN_MASK,
                                           FI_RGBA_BLUE_MASK,
                                           TRUE);
                FreeImage_Unload(src);

                {
                    QImage tmp = QImage(bits,
                                        width,
                                        height,
                                        scan_width,
                                        QImage::Format_RGB32);
                    *out = tmp;
                    out->bits(); // deep copy
                }
                free(bits);
            } else {
                qWarning() << "could not convert to 32bits: " << filename;
                *out = QImage();
            }

            // bitmap successfully loaded!
        }
    } catch (std::exception& e) {
        qCritical() << "caught exception: " << e.what();
        *out = QImage();
    }
    catch (...) {
        qCritical() << "found other error";
        *out = QImage();
    }
    return out;
}

QByteArray MiscTools::getMD5Sum(const QString & path)
{
    QFile file(path);

    if (file.open(QIODevice::ReadOnly | QIODevice::Unbuffered)) {
        QCryptographicHash md5Hash(QCryptographicHash::Md5);
        md5Hash.addData(file.readAll());
        file.close();
        return md5Hash.result();
    } else {
        return QByteArray();
    }
}


bool MiscTools::changeTimeStamp(const QString& filenameIn,
                                const QString& filenameOut,
                                const QDateTime& newDate,
                                const bool saveAsTextIfNotWOrks)
{
    // only JPEG
    QString ext = QFileInfo(filenameIn).completeSuffix().toLower();

    if (ext == "jpg" || ext == "jpeg") {
        ExifLoader *l;
        /* Create an ExifLoader object to manage the EXIF loading process */
        l = exif_loader_new();

        bool found = false;
        if (l) {
            ExifData *ed;

            /* Load the EXIF data from the image file */
            exif_loader_write_file(l, filenameIn.toLocal8Bit().data());

            /* Get a pointer to the EXIF data */
            ed = exif_loader_get_data(l);

            /* The loader is no longer needed--free it */
            exif_loader_unref(l);
            l = NULL;

            // the source file didn't have exif tags
            if (!ed) {
                ed = exif_data_new();
            }

            // next modify target time stamps exif data

            QVector<QPair<ExifTag, ExifIfd> > tags(QVector<QPair<ExifTag, ExifIfd> >()
                                                   << QPair<ExifTag, ExifIfd>(EXIF_TAG_DATE_TIME_ORIGINAL, EXIF_IFD_EXIF)
                                                   << QPair<ExifTag, ExifIfd>(EXIF_TAG_DATE_TIME_DIGITIZED, EXIF_IFD_EXIF)
                                                   << QPair<ExifTag, ExifIfd>(EXIF_TAG_DATE_TIME, EXIF_IFD_0));
            for (int i=0; i<(int)tags.size(); i++) {
                /* See if this tag exists */
                ExifEntry *entry =
                    exif_data_get_entry ( ed,
                                          tags[i].first);
                if (!entry) {
                    entry = exif_entry_new ();
                    exif_content_add_entry (ed->ifd[tags[i].second], entry);
                    exif_entry_initialize (entry, tags[i].first);
                }

                if (entry->data) free (entry->data);
                entry->components = 20;
                entry->size = sizeof (char) * entry->components;
                entry->data = (unsigned char*)malloc (entry->size);
                if (!entry->data) {
                    qCritical() << "not enough memory";
                    throw "Not enough memory.";
                }

#ifdef _MSC_VER
                const QByteArray dateStr = newDate.toString(TOOLS_TIME_FORMAT).toLocal8Bit();
                strcpy_s ((char *) entry->data, dateStr.length(), dateStr.data());
#else
                strcpy ((char *) entry->data, newDate.toString(TOOLS_TIME_FORMAT).toLocal8Bit().data());
#endif
                entry->data[19] = 0;
                found = true;
            }

            /* Parse the JPEG file from source and save it in target */
            JPEGData* jdata;
            jdata = jpeg_data_new ();
            jpeg_data_load_file (jdata, filenameIn.toLocal8Bit().data());

            jpeg_data_set_exif_data (jdata, ed);

            /* Save the modified image. */
            jpeg_data_save_file (jdata, filenameOut.toLocal8Bit().data());
            jpeg_data_unref (jdata);

            exif_data_unref (ed);
        }

        if (found) {
            return true;
        }
    }

    // saving has not worked, so write text file, if requested
    if (saveAsTextIfNotWOrks) {
        const QString textFilenameOut = filenameOut + ".txt";
        QFile file(textFilenameOut);
        file.open(QIODevice::WriteOnly | QIODevice::Text);
        QTextStream out(&file);
        out << "acquisition date: " << newDate.toString(TOOLS_TIME_FORMAT) << "\n";
        file.close();
        return true;
    }
    return false;
}

/**
 * try to find creation date. If datetimeoriginal then return immediately
 * otherwise wait for possibly other time
 * also try to determine orientation
 */
static bool ParseMetadataTime(FIBITMAP *dib,
                              FREE_IMAGE_MDMODEL model,
                              QDateTime& creationDate,
                              const QString& searchTag)
{
    FITAG *tag = NULL;
    FIMETADATA *mdhandle = NULL;

    mdhandle = FreeImage_FindFirstMetadata(model, dib, &tag);

    bool found = false;
    if(mdhandle) {
        do {
        // get the tag key
            const char *key = FreeImage_GetTagKey(tag);
        // convert the tag value to a string
            const char *value = FreeImage_TagToString(model, tag);

            QString keyAnalyse = QString(key).trimmed();
            if (searchTag.length() > 0 && keyAnalyse == searchTag) {
                creationDate = QDateTime::fromString(QString(value), TOOLS_TIME_FORMAT);
                found = true;
            } else if ( searchTag.length() == 0 && keyAnalyse.contains(QString("DateTime")) ) {
                QDateTime tmp = QDateTime::fromString(QString(value), TOOLS_TIME_FORMAT);
                if (tmp.isValid()) {
                    creationDate = tmp;
                    found = true;
                }
            }
        } while(!found && FreeImage_FindNextMetadata(mdhandle, &tag) );
    }
    FreeImage_FindCloseMetadata(mdhandle);
    return found;
}

/**
 * try to find orientation
 */
static bool ParseMetadataOrientation(FIBITMAP *dib,
                                     FREE_IMAGE_MDMODEL model,
                                     OrientationJPG& orientation)
{
    FITAG *tag = NULL;
    FIMETADATA *mdhandle = NULL;

    mdhandle = FreeImage_FindFirstMetadata(model, dib, &tag);

    bool found = false;
    if(mdhandle) {
        do {
        // get the tag key
            const char *key = FreeImage_GetTagKey(tag);
        // convert the tag value to a string
            const char *value = FreeImage_TagToString(model, tag);

            QString keyAnalyse = QString(key).trimmed();
            if (keyAnalyse.contains(QString("Orientation")) ) {
                const int val = *(char *)FreeImage_GetTagValue(tag);
                FreeImage_TagToString(model, tag);
                //const int val = QString(value).toInt();
                if (val > OrientationJPG__first &&
                        val < OrientationJPG__last) {
                    orientation = (OrientationJPG) val;
                    found = true;
                    break;
                } else {
                    qWarning() << "unknown orientation exif value" << val << ", from:" << value;
                    // otherwise assume landscape orientation
                    orientation = OrientationJPG_normal;
                }
            }
        } while(FreeImage_FindNextMetadata(mdhandle, &tag));
    }
    FreeImage_FindCloseMetadata(mdhandle);
    return found;
}


bool MiscTools::getTimeStampAndOrientation(const QString& filename,
                                           QDateTime& timeOut,
                                           OrientationJPG& orientationOut)
{
    orientationOut = OrientationJPG_normal;
    int flags = FIF_LOAD_NOPIXELS;
    FIBITMAP *dib = NULL;
    try {
    // load a file using the FIF_LOAD_NOPIXELS flag
        FREE_IMAGE_FORMAT fif = FreeImage_GetFIFFromFilename(filename.toLocal8Bit().data());
        if (FreeImage_FIFSupportsNoPixels(fif) != (BOOL)true) {
            return false;
        }
        dib = FreeImage_Load(fif, filename.toLocal8Bit().data(), flags);
        if(!dib) throw(1);
//        // check that dib does not contain pixels
//        BOOL bHasPixel = FreeImage_HasPixels(dib);
//        assert(bHasPixel == FALSE);

        // parse time
        bool foundTime = false;
        // perform two runs: first strict time searching ("OriginalDateTime", "DateTimeDigitized")
        // if not found, search for "DateTime"
        QStringList searchFor; searchFor << "DateTimeOriginal" << "DateTimeDigitized" << "";
        for (int i=0; i<searchFor.length(); i++) {
            if (ParseMetadataTime(dib, FIMD_COMMENTS, timeOut, searchFor[i])) {
            } else if (ParseMetadataTime(dib, FIMD_EXIF_MAIN, timeOut, searchFor[i])){
                foundTime = true;
            } else if (ParseMetadataTime(dib, FIMD_EXIF_EXIF, timeOut, searchFor[i])){
                foundTime = true;
            } else if (ParseMetadataTime(dib, FIMD_EXIF_GPS, timeOut, searchFor[i])){
                foundTime = true;
            } else if (ParseMetadataTime(dib, FIMD_EXIF_MAKERNOTE, timeOut, searchFor[i])){
                foundTime = true;
            } else if (ParseMetadataTime(dib, FIMD_IPTC, timeOut, searchFor[i])){
                foundTime = true;
            } else if (ParseMetadataTime(dib, FIMD_XMP, timeOut, searchFor[i])) {
                foundTime = true;
            }

            if (foundTime) {
                break;
            }
        }

        // parse orientation
        if (ParseMetadataOrientation(dib, FIMD_COMMENTS, orientationOut)) {
        } else if (ParseMetadataOrientation(dib, FIMD_EXIF_MAIN, orientationOut)){
        } else if (ParseMetadataOrientation(dib, FIMD_EXIF_EXIF, orientationOut)){
        } else if (ParseMetadataOrientation(dib, FIMD_EXIF_GPS, orientationOut)){
        } else if (ParseMetadataOrientation(dib, FIMD_EXIF_MAKERNOTE, orientationOut)){
        } else if (ParseMetadataOrientation(dib, FIMD_IPTC, orientationOut)){
        } else if (ParseMetadataOrientation(dib, FIMD_XMP, orientationOut)) {
        }

        FreeImage_Unload(dib);
        qDebug() << foundTime;
        return foundTime;
    }
    catch(int) {
        if(dib) FreeImage_Unload(dib);
        return false;
    }
}




}
// end namespace misctools
