/***********************************************************************
 * This file is part of module_misc.
 *
 * module_misc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * module_misc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with module_misc.  If not, see <http://www.gnu.org/licenses/>
 *
 *
 * Copyright (C) 2014, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#ifndef SMARTPOINTERS_H
#define SMARTPOINTERS_H

#include <memory>
#include <QImage>
#include <QPixmap>
#include <QString>
#include <QDebug>

namespace MiscTools
{

/**
 * @brief This file provides some definitions such that memory heavy data can be
 * located onto the stack, using smart pointers
 */

/**
 * @brief pQImage a smart pointer to a QImage to avoid much data on the stack.
 * The smart pointer always has an object (!)
 */
class pQImage : public std::shared_ptr<QImage>
{
public:
    /**
     * @brief pQImage enforce a creation of an empty QImage
     */
    pQImage()
        : std::shared_ptr<QImage>(new QImage())
    {
    }

    /**
     * @brief pQImage load from file
     * @param filename the file
     */
    pQImage(const QString& filename)
        : std::shared_ptr<QImage>(new QImage(filename))
    {}

    /**
     * @brief pQImage create this as reference to another pQImage
     * @param other the other smart pointer to the QImage
     */
    pQImage(const pQImage& other)
        : std::shared_ptr<QImage>()
    {
        this->reset();
        std::shared_ptr<QImage>::operator =( other);
        if (other.get() == 0) {
            std::shared_ptr<QImage>::operator =( std::make_shared<QImage>(QImage()));
        }
    }

    /**
     * @brief pQImage obtain ownership of a QImage pointer
     * @param other the pointer to an QImage -> must not be deleted somewhere else
     */
    pQImage(QImage* other)
        : std::shared_ptr<QImage>(other)
    {
    }

    /**
     * @brief pQImage a smart pointer to an image given an actual QImage object
     * @param image the object to be referenced to
     */
    pQImage(const QImage& image)
        : std::shared_ptr<QImage>(new QImage(image))
    {
    }
};

/**
 * @brief pQPixmap a smart pointer to a QPixmap to avoid much data on the stack
 * The smart pointer always has an object (!)
 */
class pQPixmap : public std::shared_ptr<QPixmap>
{
public:
    /**
     * @brief pQPixmap enforce a creation of an empty QPixmap
     */
    pQPixmap()
        : std::shared_ptr<QPixmap>(new QPixmap())
    {
    }

    /**
     * @brief pPixmap load from file
     * @param filename the file
     */
    pQPixmap(const QString& filename)
        : std::shared_ptr<QPixmap>(new QPixmap(filename))
    {}

    /**
     * @brief pQPixmap create this as reference to another pQPixmap
     * @param other the other smart pointer to the QPixmap
     */
    pQPixmap(const pQPixmap& other)
        : std::shared_ptr<QPixmap>()
    {
        this->reset();
        std::shared_ptr<QPixmap>::operator =( other);
        if (other.get() == 0) {
            std::shared_ptr<QPixmap>::operator =( std::make_shared<QPixmap>(QPixmap()));
        }
    }

    /**
     * @brief pQPixmap obtain ownership of a QPixmap pointer
     * @param other the pointer to an QPixmap -> must not be deleted somewhere else
     */
    pQPixmap(QPixmap* other)
        : std::shared_ptr<QPixmap>(other)
    {
    }

    /**
     * @brief pQPixmap a smart pointer to an image given an actual QPixmap object
     * @param image the object to be referenced to
     */
    pQPixmap(const QPixmap& image)
        : std::shared_ptr<QPixmap>(new QPixmap(image))
    {
    }
};

}
#endif // SMARTPOINTERS_H
