/***********************************************************************
 * This file is part of module_thumbnailcache.
 *
 * module_thumbnailcache is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_thumbnailcache is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_thumbnailcache.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QDebug>
#include <QDir>
#include <QSqlQuery>
#include <QSqlTableModel>
#include <QSqlRecord>
#include <QDateTime>
#include <QBuffer>
#include <QSqlError>
#include <QSvgRenderer>
#include <QPainter>
#include <QApplication>
#include <QThread>
#include <QtMultimedia/QMediaPlaylist>
#include <QtMultimedia/QVideoFrame>

#include "fastimagethumbnail.h"

// initialization of member variables
QMutex FastImageThumbnail::_databaseMutex;
QMutex FastImageThumbnail::_ioMutex;

// #define // DEBUG if (false)
FastImageThumbnail::FastImageThumbnail(const QString& location, QObject* parent )
    : QObject(parent)
    , _db(NULL)
    , _player(new QMediaPlayer())
    , _videoProbe(new QVideoProbe())
    , _playerHasMedia(false)
    , _videoWidget(new QVideoWidget())
    , _movieFrame(":/thumbnail/moviewframe.png")
    , _thumbWaitTime(100)
{
    _connectToLocalDB(location);

    _player->setVideoOutput(_videoWidget);
    //_videoWidget->setGeometry(100,100,256,256);
    _videoWidget->setVisible(false);
    _player->setMuted(true);
    //_videoWidget->show();

    connect(_videoProbe, SIGNAL(videoFrameProbed(QVideoFrame)), this, SLOT(newVideoFrame(QVideoFrame)));
    qDebug() << "probing: " << _videoProbe->setSource(_player);

    *_movieFrameThumbnail = _movieFrame->scaled(QSize(256,256), Qt::KeepAspectRatio, Qt::SmoothTransformation).convertToFormat(QImage::Format_RGB32, Qt::ImageConversionFlag::DiffuseDither | Qt::ImageConversionFlag::ThresholdAlphaDither);

    connect(&_playTimer, SIGNAL(timeout()), SLOT(playForThumbnail()));
}

FastImageThumbnail::~FastImageThumbnail()
{
    _disconnect();
    delete _db;
    _db = 0;
    //delete _player;
    //delete _player2;
    //delete _videoWidget;
}

FastImageThumbnail::Thumbnail_Type FastImageThumbnail::getThumbnail(const QFileInfo& fileinfo,
                                                                    MiscTools::pQImage &thumbnailOut,
                                                                    const bool onlyFast)
{
    // 1. try to obtain the thumbnail from the metadata
    const QString completePath = fileinfo.canonicalFilePath();
    FIBITMAP* thumbnail =  _tryLoadThumbnail(completePath);
    if (thumbnail != 0) {
        //DEBUG qDebug() << "directly loading thumbnail from image metadata";
        _convertToQImage(thumbnail, thumbnailOut);
        FreeImage_Unload(thumbnail);
        return THUMB_TYPE_IMAGE;
    }
    //FreeImage_Unload(thumbnail);
    //return false;

    // 2. try to access local thumbnail database storage
    if (_checkLocalCache(fileinfo, thumbnailOut)) {
        // DEBUG qDebug() << "loading thumbnail from local image cache";
        FreeImage_Unload(thumbnail);
        return THUMB_TYPE_IMAGE;
    }

    if (onlyFast) {
        FreeImage_Unload(thumbnail);
        return THUMB_TYPE_NO_FAST_THUMB;
    }
    // DEBUG qDebug() << "recompute thumbnail";
    FreeImage_Unload(thumbnail);

    // otherwise (re-)compute
    // a: load
    bool success = false;
    FIBITMAP* orig = _loadImage(completePath);
    if (orig != 0) {
        // b: resize
        thumbnail = _generateThumbnail(orig);
        FreeImage_Unload(orig);

        if (thumbnail != 0) {

            // c: convert format
            if (_convertToQImage(thumbnail, thumbnailOut))  {
                FreeImage_Unload(thumbnail);
                success = true;
            }
        }
    }

    // try to use Qts routines
    if (!success) {
        // now try to load as SVG
        QSvgRenderer renderer(completePath);
        if (renderer.isValid()) {
            const QSize suggestedSize = renderer.defaultSize();
            const double ratio = suggestedSize.width() / double(suggestedSize .height());
            QSize targetSize;
            if (ratio > 1) {
                targetSize.setWidth(256);
                targetSize.setHeight(256.0/ratio);
            } else {
                targetSize.setHeight(256);
                targetSize.setWidth(256.0*ratio);
            }
            *thumbnailOut = QImage(targetSize, QImage::Format_ARGB32);
            thumbnailOut->fill(Qt::white);

            QPainter painter(thumbnailOut.get());
            renderer.render(&painter);
            painter.end();
            *thumbnailOut = thumbnailOut->convertToFormat(QImage::Format_RGB32, Qt::ImageConversionFlag::DiffuseDither | Qt::ImageConversionFlag::ThresholdAlphaDither);

        }
        if (thumbnailOut->isNull() || thumbnailOut->width() == 0 || thumbnailOut->height() == 0) {
            // try to load as QImage now
            thumbnailOut->load(completePath);
            if (!(thumbnailOut->isNull() || thumbnailOut->width() == 0 || thumbnailOut->height() == 0)) {
                *thumbnailOut = thumbnailOut->scaled(QSize(256,256), Qt::KeepAspectRatio, Qt::SmoothTransformation).convertToFormat(QImage::Format_RGB32, Qt::ImageConversionFlag::DiffuseDither | Qt::ImageConversionFlag::ThresholdAlphaDither);
            } else {
                // maybe it is a video ?

                //_player2->setMedia(QUrl::fromLocalFile(completePath));
                //if (_player2->mediaStatus() != QMediaPlayer::InvalidMedia) {
                    _playlist.append(completePath);

                    //_player->setMedia(QUrl::fromLocalFile(completePath));
                    emit newTimer(); /*
                    _playTimer.stop();
                    _playTimer.setSingleShot(true);
                    _playTimer.setInterval(_thumbWaitTime);
                    _playTimer.start();*/
                    return THUMB_TYPE_TEST_VIDEO;
                //}
            }
        }

    }

    if (thumbnailOut->isNull() || thumbnailOut->width() == 0 || thumbnailOut->height() == 0) {
        return THUMB_TYPE_FAILED_LOADING;
    }

    // update cache
    _updateCache(fileinfo, thumbnailOut);

    return THUMB_TYPE_IMAGE;
}

void FastImageThumbnail::newVideoFrame(const QVideoFrame& frameOrig)
{
    QVideoFrame frame = frameOrig;
    const QImage::Format format = frame.imageFormatFromPixelFormat(frame.pixelFormat());
    if (format != QImage::QImage::Format_Invalid ) {
        if (frame.map(QAbstractVideoBuffer::ReadOnly)) {
            *_lastImage = QImage (frame.bits(), frame.size().width(), frame.size().height(), frame.bytesPerLine(), format).copy();
            frame.unmap();
            //qDebug() << "probed image " << tmp.size() << format <<  frame.pixelFormat() << frame.size() << frame.pixelFormat();
        } else {
            qDebug() << "could not map video frame to memory ";
        }
    } else {
        *_lastImage = QImage();
        qDebug() << "wrong format of video stream for thumbnail generation";
    }
}

void FastImageThumbnail::playForThumbnail()
{
    _playTimer.stop();
    if (_playerHasMedia //_player->mediaStatus() != QMediaPlayer::NoMedia
            && _player->mediaStatus() != QMediaPlayer::LoadingMedia) {
        QFileInfo fileinfo(_player->media().canonicalUrl().toLocalFile());
        if (_player->mediaStatus() == QMediaPlayer::InvalidMedia) {
            emit errorThumb(fileinfo);
            _playerHasMedia = false;
        } else {
            if (_player->state() == QMediaPlayer::PlayingState) {
                _player->pause();
                MiscTools::pQImage thumbnailOut ( _lastImage->scaled(QSize(256,256), Qt::KeepAspectRatio, Qt::SmoothTransformation).convertToFormat(QImage::Format_RGB32, Qt::ImageConversionFlag::DiffuseDither | Qt::ImageConversionFlag::ThresholdAlphaDither));

                if (thumbnailOut->isNull()  || thumbnailOut->width() == 0 || thumbnailOut->height() == 0) {
                    emit errorvideoThumb(fileinfo);
                } else {
                    QPainter painter(thumbnailOut.get());
                    painter.drawImage(0,0,_movieFrame->scaled(thumbnailOut->size(),
                                                             Qt::IgnoreAspectRatio,
                                                             Qt::SmoothTransformation));
                    painter.end();
                    _updateCache(fileinfo, thumbnailOut);
                    emit videoFrame(fileinfo, thumbnailOut);
                }
                _player->stop();
                _playerHasMedia = false;
            } else {
                _player->setPosition(0.5*_player->duration());
                _player->play();
                _playTimer.setSingleShot(true);
                _playTimer.setInterval(_thumbWaitTime);
                _playTimer.start();

            }
        }
    }


    if (!_playlist.isEmpty() && !_playerHasMedia ) { // _player->mediaStatus() == QMediaPlayer::NoMedia) {
        _player->setMedia(QUrl::fromLocalFile(_playlist.first()));
        _playerHasMedia = true;
        _playlist.pop_front();

        _playTimer.setSingleShot(true);
        _playTimer.setInterval(_thumbWaitTime);
        _playTimer.start();
    } else if (_player->mediaStatus() == QMediaPlayer::LoadingMedia) {
        _playTimer.setSingleShot(true);
        _playTimer.setInterval(_thumbWaitTime);
        _playTimer.start();
    }
}



bool FastImageThumbnail::_convertToQImage(FIBITMAP * input, MiscTools::pQImage &out)
{
    const unsigned w = FreeImage_GetWidth(input), h = FreeImage_GetHeight(input);
    *out = QImage(w, h, QImage::Format_ARGB32);
    if ((unsigned)out->width() != w || (unsigned)out->height() != h
            || out->width() == 0 || out->height() == 0) {
        return false;
    }

    FIBITMAP * bgr = FreeImage_ConvertTo24Bits(input);
    if (bgr == NULL) {
        return false;
    }
    // Calculate the number of bytes per pixel (3 for 24-bit or 4 for 32-bit)
    int bytespp = FreeImage_GetLine(bgr) / w;
    for(unsigned y = 0; y < h; y++) {
        BYTE *bits = FreeImage_GetScanLine(bgr, y);
        if (bits == 0) {
            *out = QImage();
            return false;
        }
        for(unsigned x = 0; x < w; x++) {
            out->setPixel(x, h-y-1, qRgb(bits[FI_RGBA_RED], bits[FI_RGBA_GREEN], bits[FI_RGBA_BLUE]));
            bits += bytespp;
        }
    }
    FreeImage_Unload(bgr);
    return true;
}

bool FastImageThumbnail::_connectToLocalDB(const QString& filename)
{
    QString locationPath = QDir::toNativeSeparators(filename);
    // DEBUG qDebug() << locationPath;

    QSqlDatabase::removeDatabase("qt_sql_default_connection");
    delete _db;
    _db = new QSqlDatabase;
    *_db = QSqlDatabase::addDatabase("QSQLITE");
    _db->setDatabaseName(locationPath);
    if (!_db->open()) {
        // maybe the cache is corrupted
        QDir tmp(locationPath.left(locationPath.lastIndexOf(QDir::separator())));
        if (tmp.exists(locationPath.right(-1+locationPath.length() - locationPath.lastIndexOf(QDir::separator())))) {
            // DEBUG qDebug() << "could not connect to db - now deleting file and retrying";
            QFile::remove(locationPath);
        }
        if (!_db->open()) {
            return false;
        }
    }

    {
        QSqlQuery query(*_db);
        QString queryStr = "CREATE TABLE IF NOT EXISTS imagecache"
                " ( filename varchar(32768), filetime INT, image BLOB);";
        if (!query.exec(queryStr))
        {
            // DEBUG qDebug() << "could not insert table";
        } else {
            // DEBUG qDebug() << "successful insertion of table";
        }
    }

    // DEBUG qDebug() << _db->lastError().text();

    // todo: hidden/system file attribute of thumbnails database for windows
    return true;
}


void FastImageThumbnail::_disconnect()
{
    //db.removeDatabase();
    _db->close();
    delete _db;
    _db = NULL;
    QSqlDatabase::removeDatabase("QSQLITE");
    QSqlDatabase::removeDatabase("qt_sql_default_connection");
}

bool FastImageThumbnail::_checkLocalCache(const QFileInfo& fileinfo,
                                          MiscTools::pQImage& out)
{
    _databaseMutex.lock();

    bool status = false;

    QSqlTableModel model(0, *_db);
    model.setTable("imagecache");
    model.setFilter("filename='" + fileinfo.canonicalFilePath() + "'");
    model.select();
    if (model.rowCount() > 0) {
        QSqlRecord record = model.record(0);
        QDateTime dbTime(QDateTime::fromTime_t(record.value("filetime").toInt()) );
        // last recorded modification time and real modification coincide ?
        if (dbTime == fileinfo.lastModified()) {
            QByteArray imgData = record.value("image").toByteArray();
            if (out->loadFromData(imgData)) {
                status = true;
            }
        } else {
            // DEBUG qDebug() << "file outdated, need to recompute thumbnail";
            // otherwise delete entries
            model.setFilter("filename='" + fileinfo.canonicalFilePath()  + "'");
            model.select();
            if (model.rowCount() >= 1) {
                model.removeRows(0, model.rowCount());
                model.submitAll();
            } else {
                // this branch is only debug
                // DEBUG qDebug() << "could not find db entry to remove";
            }
        }
    } else {
        // DEBUG qDebug() << fileinfo.canonicalFilePath()<< " not found in DB";
    }

    _databaseMutex.unlock();
    return status;
}

bool FastImageThumbnail::_updateCache(const QFileInfo& fileinfo,
                                      const MiscTools::pQImage& thumbnail)
{
    QByteArray ba;
    QBuffer buffer(&ba);
    buffer.open(QIODevice::WriteOnly);
    thumbnail->save(&buffer, "jpg", 85);

    _databaseMutex.lock();

    // DEBUG qDebug() << "inserting";
    QSqlTableModel model(0, *_db);
    model.setTable("imagecache");
    int row=0;
    model.insertRows(row, 1);
    // DEBUG qDebug() << fileinfo.canonicalFilePath();
    model.setData(model.index(row, 0), fileinfo.canonicalFilePath());
    model.setData(model.index(row, 1), fileinfo.lastModified().toTime_t());
    model.setData(model.index(row, 2), QVariant(ba));
    if (!model.submitAll()) {
        // DEBUG qDebug() << "error submit";
        _databaseMutex.unlock();
        return false;
    }

    _databaseMutex.unlock();
    return true;
}

FIBITMAP* FastImageThumbnail::_generateThumbnail(FIBITMAP* in)
{
    FIBITMAP* thumbnail = 0;
    // create a thumbnail and convert to a standard bitmap type
    thumbnail = FreeImage_MakeThumbnail(in, THUMBNAIL_MAX_SIZE, TRUE);
    return thumbnail;
}

FIBITMAP*  FastImageThumbnail::_tryLoadThumbnail(const QString& lpszPathName )
{
    QMutexLocker l(&_ioMutex);

    FREE_IMAGE_FORMAT fif = FIF_UNKNOWN;

    // check the file signature and deduce its format
    fif = FreeImage_GetFileType(lpszPathName.toLatin1(), 0);
    if(fif == FIF_UNKNOWN) {
        fif = FreeImage_GetFIFFromFilename(lpszPathName.toLatin1());
    }

    FIBITMAP* img = 0;
    // check that the plugin has reading capabilities ...
    if((fif != FIF_UNKNOWN) && FreeImage_FIFSupportsReading(fif)) {
        img = FreeImage_Load(fif, lpszPathName.toLatin1(), FIF_LOAD_NOPIXELS);
    }
    if (img == 0) {
        return 0;
    }

    FIBITMAP* thumbnailTmp = 0;
    // try to load thumbnail
    if (FreeImage_GetThumbnail(img) ) {
        thumbnailTmp = FreeImage_GetThumbnail(img);
    }

    FIBITMAP* thumbnail = 0;
    if (thumbnailTmp != 0) {
        thumbnail = FreeImage_Clone(thumbnailTmp);
    }
    FreeImage_Unload(img);

    return thumbnail;
}

FIBITMAP* FastImageThumbnail::_loadImage(const QString& lpszPathName)
{
    QMutexLocker l(&_ioMutex);

    FREE_IMAGE_FORMAT fif = FIF_UNKNOWN;
    // check the file signature and deduce its format
    fif = FreeImage_GetFileType(lpszPathName.toLatin1(), 0);
    if(fif == FIF_UNKNOWN) {
        fif = FreeImage_GetFIFFromFilename(lpszPathName.toLatin1());
    }

    // check that the plugin has reading capabilities ...
    if((fif != FIF_UNKNOWN) && FreeImage_FIFSupportsReading(fif)) {
        return FreeImage_Load(fif, lpszPathName.toLatin1() );
    }
    return 0;
}

