/***********************************************************************
 * This file is part of module_thumbnailcache.
 *
 * module_thumbnailcache is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_thumbnailcache is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_thumbnailcache.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#ifndef FASTIMAGETHUMBNAIL_H
#define FASTIMAGETHUMBNAIL_H

#include <QObject>
#include <QFileInfo>
#include <QMutex>
#include <QSqlDatabase>
#include <QTimer>
#include <FreeImage.h>
#include <QtGui/QMovie>
#include <QtMultimedia/QMediaPlayer>
#include <QtMultimedia/QVideoProbe>
#include <QtMultimediaWidgets/QGraphicsVideoItem>

#define THUMBNAIL_MAX_SIZE 256
#define THUMBNAIL_DATAFILE ".mythumbs.sqlite.db"

#include "smartpointers.h"

/**
 * A threadsafe class for fast obtaining of thumbnails
 *
 * The class checks for thumbnails in metadata. If not found it
 * tries to search in a local cache (SQLite database). If still not
 * found or outdated cache, it would try to (re-)compute the
 * thumbnail. All methods are threadsafe. But this has not been
 * tested veeery extensely.
 *
 * todo: windows, set hidden attribute
 *
 * @author Dominik Rueß dominik-ruess@gmx.de
 * @date 2011-09-06
 */
class FastImageThumbnail : public QObject
{
    Q_OBJECT

public:

    enum Thumbnail_Type{
        THUMB_TYPE_NO_FAST_THUMB,
        THUMB_TYPE_FAILED_LOADING,
        THUMB_TYPE_IMAGE,
        THUMB_TYPE_VIDEO,
        THUMB_TYPE_TEST_VIDEO
    };

    /**
      * init the database connection
      * @param QString location name of the sqlite database file
      */
    FastImageThumbnail(const QString& location, QObject* parent = 0);

    /**
      * destroy and disconnect
      */
    ~FastImageThumbnail();

    /**
     * get the thumbnail for a given file
     *
     * the function tries to obtain the thumbnail in the following order:
     * 1. does a thumbnail in the metadata exist ?
     * 2. does a thumbnail in the local database file exist ?
     * 2a. if yes, is it up to date ?
     * 3. (re-)compute thumbnail and save to local database file
     */
     Thumbnail_Type getThumbnail(const QFileInfo& fileinfo,
                                 MiscTools::pQImage& thumbnailOut,
                                 const bool onlyFast = false);

signals:
     void videoFrame(const QFileInfo&, const MiscTools::pQImage& thumbnailOut);
     void errorThumb(const QFileInfo&);
     void errorvideoThumb(const QFileInfo&);
     void newTimer();

public slots:
     void playForThumbnail();

private slots:
     void newVideoFrame(const QVideoFrame &frame);

private:

    /**
     * convert a FreeImage to a QImage (including flipping of y-dimension)
     */
     bool _convertToQImage(FIBITMAP * input, MiscTools::pQImage& out);

     bool _connectToLocalDB(const QString& filename);

     void _disconnect();

     bool _checkLocalCache(const QFileInfo& fileinfo,
                           MiscTools::pQImage& out);

     bool _updateCache(const QFileInfo& fileinfo,
                       const MiscTools::pQImage& thumbnail);

    /**
     * fast computation of a thumbnail
     * IMPORTANT: you have to unload the thumbnail on your own (as opposed to _tryLoadThumbnail)
     */
     FIBITMAP* _generateThumbnail(FIBITMAP* in);

    /**
     * try to load a thumbnail from an image path
     * IMPORTANT: do not unload this thumbnail with FreeImage_Unload(thumbnail)
     */
     FIBITMAP* _tryLoadThumbnail(const QString& lpszPathName);

    /**
     * load a complete image
     */
     FIBITMAP* _loadImage(const QString& lpszPathName);

    // Variables ---------------------------------------------------------------
    /**
     * make database operations atomic
     */
    static QMutex _databaseMutex;

    /**
     * make I/O operations atomic
     * However, this might not be a problem at all. But it does
     * cascade the heavy io image operations, possibly
     * making the single ones faster.
     */
    static QMutex _ioMutex;

    /**
      * holds the permanent connection to the database
      */
    QSqlDatabase* _db;


    QMediaPlayer *_player;
    QVideoProbe *_videoProbe;
    MiscTools::pQImage _lastImage;
    bool _playerHasMedia;
    QVideoWidget *_videoWidget;
    MiscTools::pQImage _movieFrame;
    MiscTools::pQImage _movieFrameThumbnail;
    QStringList _playlist;
    QTimer _playTimer;
    int _thumbWaitTime;
};

#endif // FASTIMAGETHUMBNAIL_H
