/***********************************************************************
 * This file is part of module_thumbnailcache.
 *
 * module_thumbnailcache is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * module_thumbnailcache is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with module_thumbnailcache.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include "thumbnailloader.h"

#include "fastimagethumbnail.h"

#include <QDebug>

QMutex ThumbnailLoader::pauseMutex;

QStringList ThumbnailLoader::returnRawFormats()
{
    return _rawFormats;
}


ThumbnailLoader::ThumbnailLoader(QObject *parent)
    : QThread(parent)
    , _thumbnailCache(NULL)
{
    _stopped = false;
    _fastOnly = false;

}


ThumbnailLoader::~ThumbnailLoader()
{
    _stringListMutex.lock();
    _stopped = true;
    _condition.wakeOne();
    _stringListMutex.unlock();

    wait();
    delete _thumbnailCache;
    _thumbnailCache = 0;
}

void ThumbnailLoader::setLocationPath(const QString& path)
{
    delete _thumbnailCache;
    _thumbnailCache = new FastImageThumbnail(path);
    connect(_thumbnailCache, SIGNAL(newTimer()), SLOT(newVideoTimer()), Qt::QueuedConnection);
    connect(_thumbnailCache,
            SIGNAL(videoFrame(QFileInfo,MiscTools::pQImage)),
            SIGNAL(loadedThumbnails(QFileInfo,MiscTools::pQImage)));
    connect(_thumbnailCache, SIGNAL(errorvideoThumb(QFileInfo)), SIGNAL(errorLoadingThumb(QFileInfo)));
    connect(_thumbnailCache, SIGNAL(errorThumb(QFileInfo)), SIGNAL(errorLoadingThumb(QFileInfo)));
}

void ThumbnailLoader::stop()
{
    _stringListMutex.lock();
    _stopped = true;
    _condition.wakeOne();
    _stringListMutex.unlock();

    wait();
}

/**
  * this will allow to sort for raw images
  */
bool sortRawToEnd(const ImageData &s1, const ImageData &s2)
{
    if (_rawFormats.contains(s1.getFileInfo().suffix().toUpper())
        && _rawFormats.contains(s2.getFileInfo().suffix().toUpper()) ) {
        return s1.getFileInfo().canonicalFilePath() < s2.getFileInfo().canonicalFilePath();
    } else  if (_rawFormats.contains(s1.getFileInfo().suffix().toUpper())) {
        return false;
    } else  if (_rawFormats.contains(s2.getFileInfo().suffix().toUpper())) {
        return true;
    }
    return s1.getFileInfo().canonicalFilePath() < s2.getFileInfo().canonicalFilePath();
}

void ThumbnailLoader::computeThumbnails(const QList<ImageDataPtr>& files,
                                        const bool fastOnly)
{
    if (_thumbnailCache == NULL) {
        qCritical() << "thumbnailcache not initialized";
        return;
    }
    QMutexLocker l(&_stringListMutex);
    _files.clear();
    for (int i=0; i<files.size(); i++) {
        _files.push_back(*dynamic_cast<ImageData*>(files[i].get()));
    }

    // sort the files such that expensive raw thumbnail generations are moved to the end
    qSort(_files.begin(), _files.end(), sortRawToEnd);

    _stopped = false;
    _fastOnly = fastOnly;

    if (!isRunning()) {
        start(LowPriority);
    } else {
        _condition.wakeOne();
    }
}

void ThumbnailLoader::newVideoTimer()
{
    _thumbnailCache->playForThumbnail();
}

void ThumbnailLoader::run()
{
    while (!_stopped) {
        QFileInfo fileinfo;

        // this allows other threads to stop the loading
        // process by locking here
        // it will finish the current thumbnail concurrently, though.
        pauseMutex.lock();
        pauseMutex.unlock();

        {
            QMutexLocker l(&_stringListMutex);
            if (_files.length() == 0) {
                if (_fastOnly) {
                    emit doneFast();
                    _fastOnly = false;
                }
                _condition.wait(&_stringListMutex);
                // continue: in the meantime it might have been stopped
                // or set empty
                continue;
            }
            fileinfo = _files.first().getFileInfo();
            _files.pop_front();
        }

        {
            MiscTools::pQImage thumbnail;
            FastImageThumbnail::Thumbnail_Type result = _thumbnailCache->getThumbnail(fileinfo,
                                                        thumbnail,
                                                        _fastOnly);

            if (result == FastImageThumbnail::THUMB_TYPE_IMAGE) {
                emit loadedThumbnails(fileinfo, thumbnail);
            } else if (!_fastOnly && result == FastImageThumbnail::THUMB_TYPE_FAILED_LOADING) {
                emit errorLoadingThumb(fileinfo);
            }
        }
    }
    _stopped = true;

    QMutexLocker l(&_stringListMutex);
    _files.clear();
}
