/***********************************************************************
 * This file is part of pivot.
 *
 * pivot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * pivot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with pivot.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#ifndef IMAGEDATAEXTENDED_H
#define IMAGEDATAEXTENDED_H

#include <memory>

#include <QFileInfo>
#include <QString>
#include <QDateTime>
#include <QDebug>

#include <imagedata.h>

#include "misctools.h"

#define MAX_NUM_COPY_DESTINATIONS 9

enum CopyStatus
{
    CopyStatusNotCopied,
    CopyStatusNoTime,
    CopyStatusNoAlbum,
    CopyStatusRemovedUnique,
    CopyStatusAlreadyCopied,
    CopyStatusCopyError,
    CopyStatusSuccess
};

struct ImageDataExtended : public ImageData
{
    ImageDataExtended(const QString& filename="")
        : ImageData(filename)
        , _eventId (-1)
        , _isManualTime(false)
        , _isAutomatedTime(false)
        , _isSemiAutomatedTime(false)
        , _showNoTimeOrAlbumFlag(false)
        , _albumID(-1)
        , _manualAlbum(false)
        , _copyStatus(CopyStatusNotCopied)
        , _isShifted(false)
        , _dayIsKnown(true)
        , _comment("")
        , _hasComment(false)
        , _isModificationTime(false)
    {
        _useCopyDestination.resize(MAX_NUM_COPY_DESTINATIONS);
        for (int i=0; i<(int)_useCopyDestination.size(); i++)
            _useCopyDestination[i] = true;
    }

    int getManualEventId() const { return _eventId; }
    bool getIsManualTime() const { return _isManualTime; }
    QDateTime getManualTime() const { return _manualTime; }
    QDateTime getAutomatedTime() const { return _automatedTime; }
    bool getIsAutomatedTime() const { return _isAutomatedTime; }    
    bool getIsFileModifiedTime() const { return _isModificationTime; }
    QDateTime getSemiAutomatedTime() const { return _semiAutomatedTime; }
    bool getIsSemiAutomatedTime() const { return _isSemiAutomatedTime; }
    bool getIsShiftedTime() const { return _isShifted;}
    QDateTime getTime(const bool mayUseShifted = true) const
    {
        if (_isShifted && _shiftedTime.isValid() && mayUseShifted) {
            return _shiftedTime;
        } else if (_isManualTime ) {
            return _manualTime;
        } else if (_isAutomatedTime) {
            return getAutomatedTime();
        } else if (_isSemiAutomatedTime) {
            return  getSemiAutomatedTime();
        } else {
            return QDateTime();
        }
    }
    bool getShowNoTimeOrAlbumFlag() const { return _showNoTimeOrAlbumFlag; }
    int getAlbumID() const { return _albumID; }
    bool getIsManualAlbum() const { return _manualAlbum; }
    CopyStatus getCopyStatus() const { return _copyStatus; }
    QString getMD5() const
    {
        if (_md5.length() == 0) {
            _md5 = MiscTools::MiscTools::getMD5Sum(getFileInfo().canonicalFilePath()).toHex();
        }
        return _md5;
    }
    bool getDayIsKnown() const { return _dayIsKnown; }
    QString getComment() const { return _comment; }
    bool getHasComment() const { return _hasComment; }

    void setManualEventId(const int eventId)
    {
        if (_eventId == eventId)
            return;
        _hasChanged = true;
        _eventId = eventId;
    }

    void setAutomatedTime(const QDateTime& time)
    {
        _automatedTime = time;
        _isAutomatedTime = true;
    }

    void setSemiAutomatedTime(const QDateTime& time)
    {
        _semiAutomatedTime = time;
        _isSemiAutomatedTime = true;
    }

    void setFileModifiedTime(const QDateTime& time)
    {
        _semiAutomatedTime = time;
        _isSemiAutomatedTime = true;
        _isModificationTime = true;
    }

    void setManualTime(const QDateTime& time)
    {
        if (!_isShifted && _manualTime == time && _isManualTime == true)
            return;
        _hasChanged = true;
        _isManualTime = true;
        _manualTime = time;
        _isShifted = false;
        _shiftedTime = _manualTime;
    }

    void setIsManualTime(const bool isManualTime)
    {
        if (_isManualTime == isManualTime)
            return;
        _hasChanged = true;
        _isManualTime = isManualTime;
    }

    void resetToAutomated() {
        _isManualTime = false;
        if ((_isSemiAutomatedTime && _manualTime != _semiAutomatedTime)
                || (!_isSemiAutomatedTime && _manualTime != _automatedTime)) {
            _hasChanged = true;
        }
    }

    void setShowNoTimeOrAlbumFlag(const bool flag)
    {
        if (_showNoTimeOrAlbumFlag == flag)
            return;
        _hasChanged = true;
        _showNoTimeOrAlbumFlag = flag;
    }

    void setCopyStatus(const CopyStatus status)
    {
        if (_copyStatus == status)
            return;
        _hasChanged = true;
        _copyStatus = status;
    }

    void setTimeShift(const qint64 shift)
    {

        QDateTime oldShiftedTime = _shiftedTime;
        if (_isAutomatedTime) {
            _shiftedTime = getAutomatedTime().addSecs(shift);
        } else if (_isSemiAutomatedTime) {
            _shiftedTime = getSemiAutomatedTime().addSecs(shift);
        } else if (_isManualTime){
            _shiftedTime = getManualTime().addSecs(shift);
        }
        const bool newShifted = shift == 0 ? false : true;
        if (_shiftedTime != oldShiftedTime
                || _isShifted != newShifted) {
            _hasChanged = true;
        }
        _isShifted = newShifted;
    }
    void setAlbumID(const int ID, const bool manual)
    {
        if (_albumID == ID && _manualAlbum == manual )
            return;

        // changeflag only if coming from or going to manually
        if (_manualAlbum != manual || manual) {
            _hasChanged = true;
        }
        _albumID = ID;
        _manualAlbum = manual;
    }

    void setDayIsKnown(const bool known)
    {
        if (_dayIsKnown == known)
            return;
        _hasChanged = true;
         _dayIsKnown = known;
    }

    void setComment(const QString& comment)
    {
        const QString tmp = comment.trimmed();
        if (tmp == _comment)
            return;
        // this will cause save, even if edited from given comment to none
        _hasChanged = true;
        _comment = tmp;
        _hasComment = tmp.length() > 0 ? true : false ;
    }

    inline bool operator<(const ImageDataExtended& second) const
    {
        return getTime() < second.getTime();
    }

    void toggleCopyDestinationStatus(const int i)
    {

        _useCopyDestination[i] = !_useCopyDestination[i];
        _hasChanged = true;
    }

    bool getCopyDestinationStatus(const int i) const
    {
        return _useCopyDestination[i];
    }

    void setCopyDestinationStatusAll(const quint64 useDestinations)
    {
        const QVector<bool> useCopyDestinationOld = _useCopyDestination;
        for (int j=0; j<MAX_NUM_COPY_DESTINATIONS; j++) {
#ifdef _MSC_VER
            if ( !(useDestinations & (1i64 << j) ) ) {
#else
            if ( !(useDestinations & (1 << j) ) ) {
#endif
                toggleCopyDestinationStatus(j);
            }
        }
        if (_useCopyDestination != useCopyDestinationOld)
            _hasChanged = true;
    }

    quint64 getCopyDestinationStatusAll() const
    {
        quint64 useDestinations = 0;
        for (int j=0; j< MAX_NUM_COPY_DESTINATIONS; j++) {
            if (getCopyDestinationStatus(j))
#ifdef _MSC_VER
                useDestinations += 1i64 << j;
#else
                useDestinations += 1 << j;
#endif
        }
        return useDestinations;
    }



private:
    int _eventId;
    QDateTime _automatedTime;
    QDateTime _semiAutomatedTime;
    QDateTime _shiftedTime;
    QDateTime _manualTime;
    bool _isManualTime;
    bool _isAutomatedTime;
    bool _isSemiAutomatedTime;
    bool _showNoTimeOrAlbumFlag;

    int _albumID;
    bool _manualAlbum;

    CopyStatus _copyStatus;
    mutable QString _md5;

    bool _isShifted;

    bool _dayIsKnown;

    QString _comment;
    bool _hasComment;

    QVector<bool> _useCopyDestination;

    bool _isModificationTime;
};

typedef std::shared_ptr<ImageDataExtended> ImageDataExtendedPtr;

#endif // IMAGEDATAEXTENDED_H
