/***********************************************************************
 * This file is part of pivot.
 *
 * pivot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * pivot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with pivot.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QDateTime>
#include <QDebug>

#include "imagewidget.h"
#include "ui_imagewidget.h"
#include "imageviewmodel.h"

ImageWidget::ImageWidget(const ImageDataExtended photo,
                         const int thumbnailSize,
                         const bool showAlbumView,
                         const bool oneAlbumIsSelected,
                         QWidget *parent)
    : QWidget(parent)
    , ui(new Ui::ImageWidget)
    , _photo(photo)
    , _thumbnailSize(thumbnailSize)
{
    ui->setupUi(this);

    if (!_photo.getTime().isValid()) {
        ui->labelDate->setText(tr("unknown date"));
    } else {
        QString dateText = _photo.getTime().toString(Qt::DefaultLocaleShortDate);
        ui->labelDate->setText(dateText);
    }

    displayThumbnail();

    timeFromForNew = new QAction(tr("Use this as &first image for new Album"), this);
    timeToForNew = new QAction(tr("Use this as &last image for new Album"), this);
    timeFromForTable = new QAction(tr("Use this as f&irst image for selected table album"), this);
    timeToForTable = new QAction(tr("Use this as la&st image for selected table album"), this);

    viewAllImages = new QAction(tr("View image with &all Images"), this);
    viewCurrentAlbum = new QAction(tr("View image with &current album"), this);

    timeFromForTable->setEnabled(oneAlbumIsSelected);
    timeToForTable->setEnabled(oneAlbumIsSelected);


    this->addAction(timeFromForNew);
    this->addAction(timeToForNew);
    QAction* tmp = new QAction(this);
    tmp->setSeparator(true);
    this->addAction(tmp);
    this->addAction(timeFromForTable);
    this->addAction(timeToForTable);
    QAction* tmp2 = new QAction(this);
    tmp2->setSeparator(true);
    this->addAction(tmp2);
    this->addAction(viewAllImages);
    if (showAlbumView) {
        this->addAction(viewCurrentAlbum);
    }
    this->setContextMenuPolicy(Qt::ActionsContextMenu);

    connect(timeFromForNew, SIGNAL(triggered()), this, SLOT(onTimeFromForNewTriggered()));
    connect(timeToForNew, SIGNAL(triggered()), this, SLOT(onTimeToForNewTriggered()));
    connect(timeFromForTable, SIGNAL(triggered()), this, SLOT(onTimeFromForTableTriggered()));
    connect(timeToForTable, SIGNAL(triggered()), this, SLOT(onTimeToForTableTriggered()));

    connect(viewAllImages, SIGNAL(triggered()), this, SLOT(onViewAllImages()));
    if (showAlbumView) {
        connect(viewCurrentAlbum, SIGNAL(triggered()), this, SLOT(onViewCurrentAlbum()));
    }

    const QString statustip = tr("filename '%1'; file time: %2")
            .arg(photo.getFileInfo().canonicalFilePath())
            .arg(photo.getTime().toString(Qt::DefaultLocaleShortDate));
    this->setStatusTip(statustip);
    this->setToolTip(statustip);
    this->setWhatsThis(statustip);
}

ImageWidget::~ImageWidget()
{
    delete timeFromForNew;
    timeFromForNew = 0;
    delete timeToForNew;
    timeToForNew= 0;
    delete timeFromForTable;
    timeFromForTable = 0;
    delete timeToForTable;
    timeToForTable = 0;

    delete ui;
}


void ImageWidget::updateThumbnailSize(const int thumbnailSize)
{
    _thumbnailSize = thumbnailSize;
    displayThumbnail();
}

void ImageWidget::displayThumbnail()
{
    const MiscTools::pQImage image = _photo.getThumbnail()->isNull() ? ImageViewModel::noThumb : _photo.getThumbnail();

    const int width = image->width();
    const int height = image->height();

    double newW, newH;
    if (width > height) {
        newW = _thumbnailSize;
        newH = height / (width/newW);
    } else {
        newH = _thumbnailSize;
        newW = width / (height/newH);
    }
    MiscTools::pQImage small ( image->scaled(newW, newH, Qt::KeepAspectRatio, Qt::SmoothTransformation) );

    ui->labelImage->setMinimumSize(newW, newH);
    ui->labelImage->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    ui->labelImage->setFixedSize(newW, newH);
    ui->labelImage->setPixmap(QPixmap::fromImage(*small));
    ui->labelDate->setScaledContents(true);


}

void ImageWidget::updateThumbnail(const MiscTools::pQImage& thumbnail)
{
    _photo.setThumbnail(thumbnail);
    displayThumbnail();
}

void ImageWidget::onTimeFromForNewTriggered()
{
    emit signalTimeFromForNew(_photo.getTime());
}

void ImageWidget::onTimeToForNewTriggered()
{
    emit signalTimeToForNew(_photo.getTime());
}

void ImageWidget::onTimeFromForTableTriggered()
{
    emit signalTimeFromForTable(_photo.getTime());
}

void ImageWidget::onTimeToForTableTriggered()
{
    emit signalTimeToForTable(_photo.getTime());
}

void ImageWidget::setTableSelected(const bool enabled)
{
    timeFromForTable->setEnabled(enabled);
    timeToForTable->setEnabled(enabled);
}

void ImageWidget::onViewAllImages()
{
    emit signalViewAllImages(_photo.getFileInfo().canonicalFilePath());
}

void ImageWidget::onViewCurrentAlbum()
{
    emit signalViewCurrentAlbum(_photo.getFileInfo().canonicalFilePath());
}
