/***********************************************************************
 * This file is part of pivot.
 *
 * pivot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * pivot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with pivot.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QSettings>
#include <QMessageBox>
#include <QWhatsThis>
#include <QStyle>
#include <QTimer>

#include "keepdirectoriesdialog.h"
#include "ui_keepdirectoriesdialog.h"
#include "settings.h"
#include "tools.h"


KeepDirectoriesDialog::KeepDirectoriesDialog(QWidget *parent)
    : QWidget(parent)
    , ui(new Ui::KeepDirectoriesDialog)
    , _infoHasBeenShown(false)
{
    ui->setupUi(this);


    _directoryDelegate.setDirectories(_directories);

    ui->tableWidget_dirs->setColumnCount(2);
    ui->tableWidget_dirs->setHorizontalHeaderLabels(QStringList() << tr("Source Directory") << tr("Author"));
    ui->tableWidget_dirs->setSelectionMode(QAbstractItemView::SingleSelection);
    ui->tableWidget_dirs->setSelectionBehavior (QAbstractItemView::SelectRows);

    ui->tableWidget_dirs->setItemDelegate(&_directoryDelegate);
    ui->tableWidget_dirs->horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);

    connect(ui->lineEdit, SIGNAL(returnPressed()),
            this, SLOT(on_pushButton_add_clicked()));

    ui->lineEdit->setFocus();

    loadSettings();

}

void KeepDirectoriesDialog::loadSettings()
{

    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    this->restoreGeometry(settings.value("directories_dialog_geometry").toByteArray());
    ui->tableWidget_dirs->setColumnWidth(0, settings.value("dirAssistant_column0Width", 200).toInt());
    ui->tableWidget_dirs->setColumnWidth(1, settings.value("dirAssistant_column1Width", 200).toInt());
}

void KeepDirectoriesDialog::setDirectories(const QStringList& directories)
{
    _directories = directories;
    ui->comboBox_newDir->clear();
    for (int i=0; i< _directories.size(); i++) {
        ui->comboBox_newDir->addItem(_directories[i]);
    }
    _directoryDelegate.setDirectories(_directories);
}

KeepDirectoriesDialog::~KeepDirectoriesDialog()
{
    QSettings settings(tr(IMAGE_ORGANIZATION_ORG), tr(IMAGE_ORGANIZATION_APP));
    settings.setValue("directories_dialog_geometry", this->saveGeometry());
    delete ui;
}

void KeepDirectoriesDialog::on_pushButton_add_clicked()
{
    if (_directories.size() == 0) {
        QMessageBox::information ( this,
                               tr("No sub directories"),
                               tr("There are no sub directories in the image directory you provided.\n"
                                  "You may want to skip this step now."));
        return;
    }
    if (ui->lineEdit->text().trimmed().length() == 0) {
        QMessageBox::critical( this,
                               tr("No text entered"),
                               tr("Please enter a new directory name for the selected entry \"%1\"").arg(ui->comboBox_newDir->currentText()));
        return;
    }

    // test duplicate entry
    for (int i=0; i< (int)ui->tableWidget_dirs->rowCount(); i++) {
        QString item = ui->tableWidget_dirs->item(i,0)->data(Qt::DisplayRole).toString();
        if (item == ui->comboBox_newDir->currentText()) {
            QMessageBox::critical( this,
                                   tr("Source directory already included"),
                                   tr("The source sub directory \"%1\" is already contained in the list").arg(ui->comboBox_newDir->currentText()));
            return;
        }
    }

    // otherwise add entry
    const int numRows = ui->tableWidget_dirs->rowCount();
    ui->tableWidget_dirs->setRowCount(numRows+1);
    QTableWidgetItem * from = new QTableWidgetItem(ui->comboBox_newDir->currentText());
    QTableWidgetItem * to = new QTableWidgetItem(ui->lineEdit->text());
    ui->tableWidget_dirs->setItem(numRows, 0, from);
    ui->tableWidget_dirs->setItem(numRows, 1, to);

    ui->lineEdit->setText("");
    ui->toolButton_delete->setEnabled(true);
    ui->comboBox_newDir->setCurrentIndex(qMin(ui->comboBox_newDir->count()-1,
                                              ui->comboBox_newDir->currentIndex()+1));
}

QList<QPair<QString, QString> > KeepDirectoriesDialog::getDirectoryMap()
{
    QList<QPair<QString, QString> > out;
    for (int i=0; i< (int)ui->tableWidget_dirs->rowCount(); i++) {
        QPair<QString, QString> tmp;
        tmp.first = ui->tableWidget_dirs->item(i,0)->data(Qt::DisplayRole).toString();
        tmp.second = ui->tableWidget_dirs->item(i,1)->data(Qt::DisplayRole).toString();
        out.push_back(tmp);
    }
    return out;
}

void KeepDirectoriesDialog::_on_pushButton_finish_clicked()
{
    emit doneEditing();
}

void KeepDirectoriesDialog::on_toolButton_delete_clicked()
{
    QList<QTableWidgetSelectionRange> selection = ui->tableWidget_dirs->selectedRanges();
    if (selection.size() == 0 || selection.size() > 1) {
        QMessageBox::warning(this,
                             tr("Nothing selected"),
                             tr("Please select an item to delete"));
        return;
    }
    const int row = selection[0].topRow();

    ui->tableWidget_dirs->removeRow(row);
    if (ui->tableWidget_dirs->rowCount() == 0) {
        ui->toolButton_delete->setEnabled(false);
    }
}

void KeepDirectoriesDialog::showInfo()
{
    if (!_infoHasBeenShown) {
        if (!MiscTools::MiscTools::lockDialog()) {
            QTimer::singleShot(10000, this, SLOT(showInfo()));
            return;
        }
        _infoHasBeenShown = true;
        Tools::provideInfo(this,
                           tr("Kept Sub-Directories Management"),
                           tr("<i>This is the dialog if you want to emphasize different authors"
                              " on copying your files.</i><br><br>"
                              "For instance, you may want to separate different photographers"
                              " or cameras on a directory base, e.g. for the backup copy destination.<br><br>"
                              "Here, for each sub-directory you may add a mapping source->target."
                              " You will just have to check for the respective copy tag in the settings dialog."),
                           tr("editmappingview"));
        MiscTools::MiscTools::unlockDialog();
    }
}
