/***********************************************************************
 * This file is part of pivot.
 *
 * pivot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * pivot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with pivot.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>
#include <QString>
#include <QStringList>
#include <QTimer>

#include "eventselectionassistant.h"
#include "viewdelegate.h"
#include "viewmodel.h"
#include "mainsettings.h"
#include "thumbnailloader.h"
#include "keepdirectoriesdialog.h"
#include "adapttimesdialog.h"
#include "previewdialog.h"
#include "copyassistant.h"
#include "copyfiles_helper.h"
#include "imagebutton.h"
#include "randomimagelist.h"
#include "dialogaboutpivot.h"

namespace Ui {
    class MainWindow;
}

#ifdef _WIN32
#define SETTINGS_FOLDER "pivot"
#else
#define SETTINGS_FOLDER ".pivot"
#endif
#define ALREADY_COPIED_DB "alreadycopied.sqlite.db"
#define SAVED_PROPERTIES_DB "savedAlbumDate.sqlite.db"
#define DATABASE_DATE_SETTINGS "settings"

#define CREATE_ALREADY_COPIED \
    QString("CREATE TABLE IF NOT EXISTS %1" \
    " ( filename varchar(32768), md5 varchar(256), filesize INT);") \
    .arg(DATABASE_MD5_TABLE)


#define DATE_TIME_CACHE_FORMAT "yyyy.MM.dd hh:mm:ss"
#define CREATE_SAVED_ITEMS \
    QString("CREATE TABLE IF NOT EXISTS %1" \
    " ( filename varchar(32768), albumid INT, datetime varchar(255), dateknown INT, comment varchar(1024), useCopyDestination BIGINT UNSIGNED );") \
    .arg(DATABASE_DATE_SETTINGS)

#define NEW_VERSION_LOCATION_PIVOT "http://dominik-ruess.de/pivot/currentVersion"
#define NEW_VERSION_MAGIC_NUMBER 29384729

#define WAIT_FOR_INFO_BOX 2000

// wait for x days (* -1) until providing donate info: default -27
#define DAYS_UNTIL_DONATE_HINT -27
#define WAIT_FOR_DONATE_HINT_MS 600000

#define WAIT_FOR_CHECK_FOR_UPDATE 120000

class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(QWidget *parent = 0);
    ~MainWindow();

public slots:
    void loadDirectory(const QString& imageDirOrig);

private slots:
    void restoreGeometryNow();
    void onloadedThumbnails(const QFileInfo& file,
                            const MiscTools::pQImage &thumbnail);
    void onErrorThumb(const QFileInfo& file);
    void onDoneFastThumbnails();

    void _onSettingsDone();

    void _loadDirectory();

    void _findImageFiles();
    void onErrorMessage(const QString& headline, const QString& errorMsg);

    void _onManageAlbums();

    void _onViewAll();

    void _onViewSelected();

    void _onViewDir();

    void _onSubDirMap();

    void _onAdaptTimes();

    void _contextMenuEvent(const QPoint &pos);

    void _onNewAlbums();

    void _doneAlbums();

    void _doneDirectoryAssistant();

    void on_actionPreview_copy_results_triggered();

    void on_actionStart_copying_triggered();

    void on_actionAbout_Qt_triggered();

    void on_actionAbout_FreeImage_triggered();

    void on_actionAbout_libexif_triggered();

    void on_actionAbout_libjpg_triggered();

    void on_actionAbout_pivot_triggered();

    void on_actionRescan_current_directory_triggered();

    void on_actionSupport_by_donation_triggered();

    void on_action_Help_triggered();

    void _resetSettings();

    void _removeUserData();

    void _saveAlbumDate( const bool dispMessage=true );

    void on_toolButton_whatsThis_clicked();

    void _showInfoBox();

    void _on_albumReset();
    void  _on_albumUp();
    void  _on_albumDown();

    void  _on_timeReset();
    void  _on_timeUp();
    void  _on_timeDown();

    void  _on_destinationsReset();
    void  _on_destinationsUp();
    void  _on_destinationsDown();

    void  _on_selectAll();
    void  _on_deselectAll();

    void _setDoneAssistant(MiscTools::ImageButton*);
    void _resetAssistant();

    void on_pushButton_assAdvanced_toggled(bool checked);
    void _on_reviewFiles_clicked();
    void _on_viewImages_clicked();
    void _on_adaptTimes_clicked();

    void on_actionCheck_for_new_version_triggered();

    void on_tabWidget_currentChanged(int index);

private:
    HandleExistingItems _getExistingFilesChoice();

    bool _connectToLocalDB(const QString& locationPath,
                           const QString& createString,
                           QSqlDatabase** db);
    void _disconnectDB(QSqlDatabase** db);

    void _getEditMenu(QMenu** menu, const bool showView);

    void closeEvent(QCloseEvent *);

    QList<ImageDataPtr> _getTimeSorted(int& currIndex,
                                         const QList<ImageDataPtr> &currentList,
                                         QVector<int> indexes = QVector<int>());
    Ui::MainWindow *ui;

    EventSelectionAssistant _eventAssistant;

    ViewDelegate _delegate;

    ViewModel _currentFiles;

    MainSettings _settings;

    QString _imageDir;

    ThumbnailLoader* _loader;

    QString _databaseLocation;
    QString _settingsLocation;
    QString _alreadyCopiedLocation;
    QString _savedAlbumDateLocation;

    QAction *_viewSelected ;
    QAction *_viewDir ;
    QAction *_viewAll ;

    QAction * _albumReset;
    QAction * _albumUp;
    QAction * _albumDown;

    QAction * _timeReset;
    QAction * _timeUp;
    QAction * _timeDown;

    QAction * _destinationsReset;
    QAction * _destinationsUp;
    QAction * _destinationsDown;

    QAction * _checkAll;
    QAction * _uncheckAll;
    QAction * _invertCheck;

    QAction * _selectAll;
    QAction * _deselectAll;

    QAction * _toggleError;

    KeepDirectoriesDialog _directoryAssistant;

    QStringList _subDirectories;

    AdaptTimesDialog _adaptTimeAssistant;

    PreviewDialog _preview;

    /**
      * holds the permanent connection to the database (copy status)
      */
    QSqlDatabase* _db;


    /**
      * holds the permanent connection to the database (previously set album/time)
      */
    QSqlDatabase* _dbFotoSettings;

    QString _lastImageDirUsed;

    QTimer _autosaveTimer;

    int _infoBoxMode;

    int _lastRow;

    MiscTools::ImageButton* _assistantButtonLoadDirectory;
    MiscTools::ImageButton* _assistantButtonViewImages;
    MiscTools::ImageButton* _assistantButtonReviewFiles;
    MiscTools::ImageButton* _assistantButtonAdaptTimes;
    MiscTools::ImageButton* _assistantButtonAlbums;
    MiscTools::ImageButton* _assistantButtonKeepDirectories;
    //ImageButton* _assistantButtonSettings;
    MiscTools::ImageButton* _assistantButtonPreview;
    MiscTools::ImageButton* _assistantButtonStart;
    QList<MiscTools::ImageButton*> _assistantButtons;

    const int _assistantButtonMinHeight;
    const int _assistantButtonMargin;

    RandomImageList* _randomImageList;

    QWidget* _currentTab;

    DialogAboutPivot _about;
};

#endif // MAINWINDOW_H

