/***********************************************************************
 * This file is part of pivot.
 *
 * pivot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * pivot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with pivot.  If not, see <http://www.gnu.org/licenses/>
 * 
 * 
 * Copyright (C) 2013, Dominik Rueß; info@dominik-ruess.de
 **********************************************************************/

#include <QDateTime>
#include <QDirIterator>
#include <QDebug>
#include <QFileInfo>
#include <QApplication>
#include <cassert>
#include <iostream>
#include <QSqlTableModel>
#include <QSqlRecord>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QCheckBox>
#include <QSettings>
#include <QPushButton>

#include "tools.h"
#include "FreeImage.h"
#include <misctools.h>
#include "copyfiles.h"
#include "imagedata.h"
#include "settings.h"

inline bool isSameBaseName(const QString& n1, const QString& n2)
{
    if (n1.length() != n2.length())
        return false;
    for (int i=0; i<(int)n1.length(); i++) {
        if (n1[i] != n2[i] &&
                !(n1[i].isDigit() && n2[i].isDigit() ) ) {
            return false;
        }
    }
    return true;
}


QList<ImageDataPtr> Tools::findFiles(const QString& dir,
                                     const bool tryFindAlreadyCopied,
                                     const QSqlDatabase* db,
                                     const QDir::Filters& filter,
                                     ProgressDialog* progress)
{
    bool newProgress = false;
    if (progress == NULL) {
        progress = new ProgressDialog;
        progress->setMinMax(0, 0);
        progress->setLabel("");
        progress->setValue(0);
        progress->show();
        //progress->setTitle(qApp->translate("findfiles", "searching for images and videos..."));
        qApp->processEvents();
        newProgress = true;
    }

    QList<ImageDataPtr> out;
    QDir scan(dir);
    QFileInfoList list = scan.entryInfoList(filter);
    for (int i=0; i<list.size(); i++) {
        if (progress->isCanceled()) {
            out.clear();
            break;
        }
        if (list.at(i).fileName() == "."
                || list.at(i).fileName() == "..") {
            continue;
        }

        if (progress->checkUpdateWanted()) {
            progress->setLabel(qApp->translate("findfiles", "Searching: ") + list.at(i).canonicalFilePath());
        }
        qApp->processEvents();

        if (list.at(i).isDir()  ) {
            out.append(Tools::findFiles(list.at(i).canonicalFilePath(), tryFindAlreadyCopied, db, filter, progress));
        } else {
            ImageDataExtendedPtr info ( new ImageDataExtended(list.at(i).canonicalFilePath()) );

            if (tryFindAlreadyCopied && db != NULL) {
                    if (CopyFiles::checkHasBeenCopied(info.get(), db)) {
                        info->setCopyStatus(CopyStatusAlreadyCopied);
                    }
            }

            QDateTime timestamp;
            MiscTools::OrientationJPG orientation = MiscTools::OrientationJPG_normal;
            if (MiscTools::MiscTools::getTimeStampAndOrientation(
                        list.at(i).canonicalFilePath(),
                        timestamp,
                        orientation) ) {
                info->setAutomatedTime(timestamp);
            }
            info->setOrientation(orientation);
            out.push_back(info);
        }
    }

    if (newProgress) {
        delete progress;
        progress = 0;
    }

    return out;
}

/** returns an output where only certain files are allowed
 and where files without timestamp are found
 */
QList<ImageDataPtr> Tools::sortItems(const QList<ImageDataPtr>& input,
                                     const QStringList& extensionsOrig,                                     
                                     const bool tryToUseMetadata,
                                     const bool tryToUseBasename,
                                     const bool tryToUseNeighbours,
                                     const bool tryToUseFileTime,
                                     /*const QSqlDatabase* db,
                                     const QString& table,
                                     const QString& cacheFormat,
                                     const bool useSavedData,
                                     const QVector<PhotoAlbum>& currAlbums,*/
                                     QList<ImageDataPtr> & discarded,
                                     QList<QDir>& resultDirs)
{
    ProgressDialog* progress = new ProgressDialog();
    progress->setModal(true);
    progress->setMinMax(1, input.size());
    progress->setValue(1);
    progress->show();
    progress->setLabel(qApp->translate("sortingafterloading", "Sorting files ...") );

    QList<ImageDataPtr> out;

    if (extensionsOrig.size() == 0) {
        discarded = input;
        return QList<ImageDataPtr> ();
    }

    QStringList extensions;
    for (int i=0; i< extensionsOrig.size(); i++) {
        extensions.push_back( extensionsOrig[i].toLower() );
    }

    resultDirs.clear();
    discarded.clear();

    // sorted inserting
    // this needs to be done on a filepath based sorting, since neighbours
    // may need to check for times etc based on the file name numbering
    QList<ImageDataPtr> sortedFilenames;
    for (int i=0; i<(int)input.size(); i++) {
        if (progress->checkUpdateWanted()) {
            progress->setValue(i);
        }
        if (progress->isCanceled()) {
            return QList<ImageDataPtr> ();
        }
        bool inserted = false;
        const int s2=(int)sortedFilenames.size();
        for (int j=0; j<s2 ; j++) {
            if (QString::compare( input[i]->getFileInfo().canonicalFilePath(),
                                  sortedFilenames[j]->getFileInfo().canonicalFilePath() ) < 0) {
                inserted = true;
                sortedFilenames.insert(j,input[i]);
                break;
            }
        }

        if (!inserted) {
            sortedFilenames.push_back(input[i]);
        }
    }
    delete progress;
    progress = 0;

    /*
    // load saved albums / times
    if (useSavedData) {
        for (int i=0; i<sortedFilenames.size(); i++) {
            ImageDataExtended* fj = static_cast<ImageDataExtended*>(sortedFilenames[i].get());

            int albumId = -1;
            QString time;
            bool knownDay = true;
            QString comment;

            // does file have a value ?

            QSqlTableModel model(0, *db);
            model.setTable(table);
            model.setFilter(QString("filename='%1'").arg(fj->getFileInfo().canonicalFilePath()));
            model.select();
            if (model.rowCount() > 0 ) {
                QSqlRecord record = model.record(0);
                time = record.value("datetime").toString();
                albumId = record.value("albumid").toLongLong();
                knownDay = record.value("dateknown").toInt() > 0;
                comment = record.value("comment").toString();

                if (comment.length() > 0) {
                    fj->setComment(comment);
                }

                if (time.length() > 0) {
                    QDateTime savedDate = QDateTime::fromString(time, cacheFormat);
                    fj->setManualTime(savedDate);
                }

                fj->setDayIsKnown(knownDay);

                // is it automated album or not ?
                bool isManualAlbum = false;

                for (int j=0; j<currAlbums.size(); j++) {
                    if (fj->getTime().isValid()
                            && currAlbums[j].from <= fj->getTime()
                            && currAlbums[j].to > fj->getTime()) {
                        isManualAlbum =  currAlbums[j].ID != albumId;
                    }
                }
                if (albumId != -1) {
                    fj->setAlbumID(albumId, isManualAlbum);
                }
            }
        }
    }
    */

    progress = new ProgressDialog();
    progress->setModal(true);
    progress->setMinMax(1, sortedFilenames.size());
    progress->setValue(1);
    progress->show();
    progress->setLabel(qApp->translate("timedetermination", "Determining file creation times ... ") );

    for (int i=0; i<(int)sortedFilenames.size(); i++) {
        if (progress->checkUpdateWanted()) {
            progress->setValue(i);
        }
        if (progress->isCanceled()) {
            sortedFilenames.clear();
            out.clear();
            break;
        }

        if (extensions.indexOf(sortedFilenames[i]->getFileInfo().suffix().toLower() ) == -1) {
            discarded.push_back(sortedFilenames[i]);
            continue;
        }


        bool found = true;
        ImageDataExtended* fi = dynamic_cast<ImageDataExtended*>(sortedFilenames[i].get());
        if (!tryToUseMetadata
                || (!fi->getIsAutomatedTime() && !fi->getIsSemiAutomatedTime() )) {
            // find a timestamp in different kind of ways
            found = false;

            if (tryToUseBasename) {
                // first of all search for files with same basename
                for (int j=0; j<(int)sortedFilenames.size(); j++) {
                    ImageDataExtended* fj = dynamic_cast<ImageDataExtended*>(sortedFilenames[j].get());
                    if ( //(fj->getIsAutomatedTime() || fj->getIsSemiAutomatedTime() )
                         (fj->getTime().isValid())
                         && fj->getFileInfo().canonicalPath() == fi->getFileInfo().canonicalPath()
                         && fj->getFileInfo().completeBaseName() == fi->getFileInfo().completeBaseName() )
                    {
                        found = true;
                        fi->setSemiAutomatedTime( fj->getTime() );
                        if (!fi->getHasComment()) fi->setComment( fj->getComment());
                        if (!fi->getIsManualAlbum()) fi->setAlbumID( fj->getAlbumID(), fj->getIsManualAlbum());
                        if (fi->getDayIsKnown()) fi->setDayIsKnown( fj->getDayIsKnown() );
                        *sortedFilenames[i] = *fi;
                        break;
                    }
                }
            }

            if (!found && tryToUseNeighbours) {
                /* search files which are ahead
                make sure that:
                - the neighbour has a automated time or semiautomated time
                - the neighbour has not its file modification time, then the current file
                  can also use its own mod time later on
                - the basename has to be the same except for digits (otherwise try modification time)
                    */
                for (int j=i-1; j>=0; j--) {
                    ImageDataExtended* fj = dynamic_cast<ImageDataExtended*>(sortedFilenames[j].get());
                    if ( (fj->getIsAutomatedTime() || fj->getIsSemiAutomatedTime() )
                         && !fj->getIsFileModifiedTime()
                         && fj->getFileInfo().canonicalPath() == fi->getFileInfo().canonicalPath()
                         && isSameBaseName(fj->getFileInfo().baseName(), fi->getFileInfo().baseName()))
                    {
                        found = true;
                        fi->setSemiAutomatedTime( fj->getTime() );
                        if (!fi->getHasComment()) fi->setComment( fj->getComment());
                        if (!fi->getIsManualAlbum()) fi->setAlbumID( fj->getAlbumID(), fj->getIsManualAlbum());
                        if (fi->getDayIsKnown()) fi->setDayIsKnown( fj->getDayIsKnown() );
                        *sortedFilenames[i] = *fi;
                        break;
                    }
                }

                if (!found) {
                    /* search for following files, same as above */
                    for (int j=i+1; j<(int)sortedFilenames.size(); j++) {
                        ImageDataExtended* fj = dynamic_cast<ImageDataExtended*>(sortedFilenames[j].get());
                        if ( (fj->getIsAutomatedTime() || fj->getIsSemiAutomatedTime() )
                             && !fj->getIsFileModifiedTime()
                             && fj->getFileInfo().canonicalPath() == fi->getFileInfo().canonicalPath()
                             && isSameBaseName(fj->getFileInfo().completeBaseName(), fi->getFileInfo().completeBaseName()))
                        {
                            found = true;
                            fi->setSemiAutomatedTime( fj->getTime() );
                            if (!fi->getHasComment()) fi->setComment( fj->getComment());
                            if (!fi->getIsManualAlbum()) fi->setAlbumID( fj->getAlbumID(),fj->getIsManualAlbum());
                            if (fi->getDayIsKnown()) fi->setDayIsKnown( fj->getDayIsKnown() );
                            *sortedFilenames[i] = *fi;
                            break;
                        }
                    }
                }
            } // end try to use neighbours


            if (!found && tryToUseFileTime) {
                fi->setFileModifiedTime( fi->getFileInfo().lastModified() );
                *sortedFilenames[i] = *fi;
            }

        }

        // now add them to output
        bool inDir = false;
        for (int k=0; k<(int)resultDirs.size(); k++) {
            if (resultDirs[k].canonicalPath() == fi->getFileInfo().canonicalPath()) {
                inDir = true;
                break;
            }
        }
        if (!inDir) resultDirs.push_back(QDir(fi->getFileInfo().absoluteDir()));
        out.push_back(sortedFilenames[i]);
    }

    delete progress;
    progress = 0;
    //qSort(out);
    return out;
}


void Tools::provideInfo(QWidget *parent,
                        const QString &headline,
                        const QString &text,
                        const QString& settingsValue,
                        const QString& showAgain,
                        const int width,
                        const int minWidth)
{
    QSettings settings(qApp->translate("tools", IMAGE_ORGANIZATION_ORG), qApp->translate("tools", IMAGE_ORGANIZATION_APP));
    if (!settings.value(settingsValue, true).toBool()) {
        return;
    }

    const int iconsize = 128;

    QDialog* dialog = new QDialog(parent, Qt::Dialog);
    dialog->setMinimumWidth(minWidth);
    dialog->setMaximumWidth(width);
    dialog->setModal(true);
    dialog->setWindowTitle(headline);
    QVBoxLayout* layout = new QVBoxLayout;
    QHBoxLayout* messageInfo = new QHBoxLayout;

    QIcon icon;
    if (QIcon::hasThemeIcon("info")) {
        icon = QIcon::fromTheme("info");
    } else {
        icon = QIcon(":/images/info.png");
    }
    QLabel* infoIcon = new QLabel;
    infoIcon->setScaledContents(false);
    MiscTools::pQPixmap pixmap (icon.pixmap(256,
                                 QIcon::Normal,
                                 QIcon::On));
    infoIcon->setPixmap(pixmap->scaled(QSize(iconsize, iconsize),
                                      Qt::KeepAspectRatio,
                                      Qt::SmoothTransformation));
    infoIcon->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);

    QLabel* info = new QLabel(text);
    info->setOpenExternalLinks(true);
    info->setTextFormat(Qt::RichText);
    info->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::MinimumExpanding);
    info->setWordWrap(true);
#ifndef _WIN32	
    info->setMargin(5);
#endif	
    info->setMaximumWidth(width-iconsize);
    messageInfo->addWidget(infoIcon);
    messageInfo->addWidget(info);

    QHBoxLayout* layoutCheckBox = new QHBoxLayout;
    layoutCheckBox->addStretch();
    QCheckBox* showAgainBox = new QCheckBox(showAgain);
    showAgainBox->setChecked(true);
    showAgainBox->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    layoutCheckBox->addWidget(showAgainBox);

    QHBoxLayout* layoutButton = new QHBoxLayout;
    layoutButton->addStretch();

    QPushButton* okButton = new QPushButton(qApp->translate("tools", "OK"));
    layoutButton->addWidget(okButton);
#ifndef _WIN32	
    layoutButton->setMargin(5);
#endif	
    qApp->connect(okButton, SIGNAL(clicked()), dialog, SLOT(close()));


    layout->addLayout(messageInfo);
    layout->addLayout(layoutCheckBox);
    layout->addLayout(layoutButton);
    dialog->setLayout(layout);

    dialog->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);

    dialog->exec();

    settings.setValue(settingsValue, showAgainBox->isChecked());

    delete dialog;
    dialog = 0;
}
